import itertools
import os
import time
from collections import defaultdict
from io import BytesIO, StringIO
from pathlib import Path
from stat import S_ISDIR
from typing import Any, Dict, List

import pandas as pd
import paramiko
import prefect

from viadot.sources.base import Source

from ..exceptions import CredentialError


class SftpConnector(Source):
    def __init__(
        self,
        file_name: str = None,
        credentials_sftp: Dict[str, Any] = None,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        STFP connector which allows for download data files, listing and downloading into Data Frame.

        Args:
            file_name (str, optional): File name to download. Defaults to None.
            credentials_sftp (Dict[str, Any], optional): SFTP server credentials. Defaults to None.

        Notes:
            self.conn is paramiko.SFTPClient.from_transport method that contains additional methods like
            get, put, open etc. Some of them were not implemented in that class.
            For more check documentation (https://docs.paramiko.org/en/stable/api/sftp.html)

            sftp = SftpConnector()
            sftp.conn.open(filename='folder_a/my_file.zip', mode='r')


        Raises:
            CredentialError: If credentials are not provided in local_config or directly as a parameter.
        """

        self.logger = prefect.context.get("logger")

        self.credentials_sftp = credentials_sftp

        if self.credentials_sftp is None:
            raise CredentialError("Credentials not found.")

        self.file_name = file_name
        self.conn = None
        self.hostname = self.credentials_sftp.get("HOSTNAME")
        self.username = self.credentials_sftp.get("USERNAME")
        self.password = self.credentials_sftp.get("PASSWORD")
        self.port = self.credentials_sftp.get("PORT")
        self.rsa_key = self.credentials_sftp.get("RSA_KEY")

        self.recursive_files_list = {}
        self.file_name_list = []
        self.recursive_files = []

    def get_conn(self) -> paramiko.SFTPClient:
        """Returns a SFTP connection object.

        Returns: paramiko.SFTPClient.
        """

        ssh = paramiko.SSHClient()

        if self.conn is None:
            if self.rsa_key is None:
                transport = paramiko.Transport((self.hostname, self.port))
                transport.connect(None, self.username, self.password)

                self.conn = paramiko.SFTPClient.from_transport(transport)

            else:
                keyfile = StringIO(self.rsa_key)
                mykey = paramiko.RSAKey.from_private_key(keyfile)
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                ssh.connect(self.hostname, username=self.username, pkey=mykey)
                time.sleep(1)
                self.conn = ssh.open_sftp()

            return self.conn

    def get_cwd(self) -> str:
        """Return the current working directory for SFTP session.

        Returns:
            str: current working directory.
        """
        return self.conn.getcwd()

    def getfo_file(self, file_name: str) -> BytesIO:
        """Copy a remote file from the SFTP server and write to a file-like object.

        Args:
            file_name (str, optional): File name to copy.

        Returns:
            BytesIO: file-like object.
        """
        flo = BytesIO()
        try:
            self.conn.getfo(file_name, flo)
            return flo
        except Exception as e:
            self.logger.info(e)

    def to_df(
        self, file_name: str, sep: str = "\t", columns: List[str] = None, **kwargs
    ) -> pd.DataFrame:
        """Copy a remote file from the SFTP server and write it to Pandas dataframe.

        Args:
            file_name (str, optional): File name to download.
            sep (str, optional): The delimiter for the source file. Defaults to "\t".
            columns (List[str], optional): List of columns to select from file. Defaults to None.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """
        byte_file = self.getfo_file(file_name=file_name)
        byte_file.seek(0)

        if Path(file_name).suffix == ".csv":
            df = pd.read_csv(byte_file, sep=sep, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".parquet":
            df = pd.read_parquet(byte_file, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".tsv":
            df = pd.read_csv(byte_file, sep="\t", usecols=columns, **kwargs)

        elif Path(file_name).suffix in [".xls", ".xlsx", ".xlsm"]:
            df = pd.read_excel(byte_file, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".json":
            df = pd.read_json(byte_file, **kwargs)

        elif Path(file_name).suffix == ".pkl":
            df = pd.read_pickle(byte_file, **kwargs)

        elif Path(file_name).suffix == ".sql":
            df = pd.read_sql(byte_file, **kwargs)

        elif Path(file_name).suffix == ".hdf":
            df = pd.read_hdf(byte_file, **kwargs)

        else:
            raise ValueError(
                f"Not able to read the file {Path(file_name).name}, unsupported filetype: {Path(file_name).suffix}"
            )

        return df

    def get_exported_files(self) -> List[str]:
        """List only exported files in current working directory.

        Returns:
            List: List of exported files.
        """
        self.file_name_list.clear()

        directory_structure = self.conn.listdir()
        for file in directory_structure:
            if "##exported.tsv" in file:
                self.file_name_list.append(file)

        return self.file_name_list

    def list_directory(self, path: str = None) -> List[str]:
        """Returns a list of files on the remote system.

        Args:
            path (str, optional): full path to the remote directory to list. Defaults to None.

        Returns:
            List: List of files.

        """

        if path is None:
            files = self.conn.listdir()
        else:
            files = self.conn.listdir(path)
        return files

    def recursive_listdir(self, path=".", files=None) -> defaultdict(list):
        """Recursively returns a defaultdict of files on the remote system.

        Args:
            path (str, optional): full path to the remote directory to list. Defaults to None.
            files (any, optional): parameter for calling function recursively.
        Returns:
            defaultdict(list): List of files.

        """

        if files is None:
            files = defaultdict(list)

        # loop over list of SFTPAttributes (files with modes)
        for attr in self.conn.listdir_attr(path):
            if S_ISDIR(attr.st_mode):
                # If the file is a directory, recurse it
                self.recursive_listdir(os.path.join(path, attr.filename), files)

            else:
                #  if the file is a file, add it to our dict
                files[path].append(attr.filename)

        self.recursive_files = files
        return files

    def process_defaultdict(self, defaultdict: any = None) -> list:
        """Process defaultdict to list of files.

        Args:
            defaultdict (any, optional): defaultdict of recursive files. Defaults to None.

        Returns:
            List: list of files.
        """
        path_list = []
        if defaultdict is None:
            defaultdict = self.recursive_files

        for item in list(defaultdict.items()):
            tuple_list_path = list(itertools.product([item[0]], item[1]))
            path_list.extend(
                [os.path.join(*tuple_path) for tuple_path in tuple_list_path]
            )
        return path_list

    def close_conn(self) -> None:
        """Closes the connection"""
        if self.conn is not None:
            self.conn.close()
            self.conn = None



from viadot.config import local_config
from viadot.flows import CloudForCustomersReportToADLS


def test_cloud_for_customers_report_to_adls():
    credentials = local_config.get("CLOUD_FOR_CUSTOMERS")
    credentials_prod = credentials["Prod"]
    channels = ["VEL_B_AFS", "VEL_B_ASA"]
    month = ["01"]
    year = ["2021"]
    flow = CloudForCustomersReportToADLS(
        report_url=credentials_prod["server"],
        env="Prod",
        channels=channels,
        months=month,
        years=year,
        name="test_c4c_report_to_adls",
        local_file_path=f"test_c4c_report_to_adls.csv",
        adls_sp_credentials_secret=credentials["adls_sp_credentials_secret"],
        adls_dir_path=credentials["adls_dir_path"],
    )
    number_of_urls = len(month) * len(year) * len(channels)
    assert len(flow.report_urls_with_filters) == number_of_urls

    result = flow.run()
    assert result.is_successful()

    task_results = result.result.values()
    assert all([task_result.is_successful() for task_result in task_results])



import os

from viadot.config import local_config

try:
    import pyrfc
except ModuleNotFoundError:
    raise

from viadot.flows import SAPToDuckDB

sap_test_creds = local_config.get("SAP").get("TEST")
duckdb_creds = {"database": "test1.duckdb"}


def test_sap_to_duckdb():
    flow = SAPToDuckDB(
        name="SAPToDuckDB flow test",
        query="""
        select
        ,CLIENT as client
        ,KNUMV as number_of_the_document_condition
        ,KPOSN as condition_item_number
        ,STUNR as step_number
        ,KAPPL as application
        from PRCD_ELEMENTS
        where KNUMV = '2003393196'
            and KPOSN = '000001'
            or STUNR = '570'
            and CLIENT = '009'
        limit 3
        """,
        schema="main",
        table="test",
        local_file_path="local.parquet",
        table_if_exists="replace",
        sap_credentials=sap_test_creds,
        duckdb_credentials=duckdb_creds,
    )

    result = flow.run()
    assert result.is_successful()

    task_results = result.result.values()
    assert all([task_result.is_successful() for task_result in task_results])

    os.remove("test1.duckdb")



import os
from pathlib import Path
from typing import Any, Dict, List

import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, BigQueryToDF

logger = logging.get_logger(__name__)


class BigQueryToADLS(Flow):
    def __init__(
        self,
        name: str = None,
        query: str = None,
        dataset_name: str = None,
        table_name: str = None,
        date_column_name: str = "date",
        start_date: str = None,
        end_date: str = None,
        credentials_key: str = "BIGQUERY",
        vault_name: str = None,
        output_file_extension: str = ".parquet",
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from BigQuery project to a local CSV or Parquet file using
        Bigquery API, then uploading it to Azure Data Lake.
        For querying on database user have to provide `query` parameter OR `dataset` and `table_name`.

        If user does not provide `query` there are 3 cases:
            If start_date and end_date are not None - all data from the start date to the end date will be retrieved.
            If start_date and end_date are left as default (None) - the data is pulled till "yesterday" (current date -1)
            If the column that looks like a date does not exist in the table, get all the data from the table.

        Args:
            name (str): The name of the flow.
            query(str, optional): SQL query to querying data in BigQuery. Format of basic query: (SELECT * FROM `{project}.{dataset_name}.{table_name}`)
                Defaults to None.
            dataset_name (str, optional): Dataset name. Defaults to None.
            table_name (str, optional): Table name. Defaults to None.
            date_column_name (str, optional): The query is based on a date, the user can provide the name
            of the date columnn if it is different than "date". If the user-specified column does not exist,
            all data will be retrieved from the table. Defaults to "date".
            start_date (str, optional): A query parameter to pass start date e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A query parameter to pass end date e.g. "2022-01-01". Defaults to None.
            credentials_key (str, optional): Credential key to dictionary where details are stored - Azure Key Vault or local config.
                credentials can be generated as key for User Principal inside a BigQuery project. Defaults to "BIGQUERY".
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of.csv for data
            which is not easy to handle with parquet. Defaults to ".parquet".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # BigQueryToDF
        self.query = query
        self.dataset_name = dataset_name
        self.table_name = table_name
        self.start_date = start_date
        self.end_date = end_date
        self.date_column_name = date_column_name
        self.vault_name = vault_name
        self.credentials_key = credentials_key

        # AzureDataLakeUpload
        self.overwrite = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.now = str(pendulum.now("utc"))
        self.timeout = timeout

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        bigquery_to_df_task = BigQueryToDF(timeout=self.timeout)
        df = bigquery_to_df_task.bind(
            query=self.query,
            dataset_name=self.dataset_name,
            table_name=self.table_name,
            credentials_key=self.credentials_key,
            start_date=self.start_date,
            end_date=self.end_date,
            date_column_name=self.date_column_name,
            vault_name=self.vault_name,
            flow=self,
        )

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)
        df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
            df_with_metadata, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_with_metadata.set_upstream(df, flow=self)
        dtypes_dict.set_upstream(df_with_metadata, flow=self)
        df_to_be_loaded.set_upstream(dtypes_dict, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



from typing import Any, Dict, List

from prefect import Flow

from viadot.tasks.azure_sql import AzureSQLDBQuery


class AzureSQLTransform(Flow):
    def __init__(
        self,
        name: str,
        query: str,
        sqldb_credentials_secret: str = None,
        vault_name: str = None,
        tags: List[str] = ["transform"],
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for running SQL queries on top of Azure SQL Database.

        Args:
            name (str): The name of the flow.
            query (str, required): The query to execute on the database.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary
            with SQL db credentials (server, db_name, user, and password).
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            tags (list, optional): Tag for marking flow. Defaults to "transform".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.query = query
        self.tags = tags
        self.sqldb_credentials_secret = sqldb_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    def gen_flow(self) -> Flow:
        query_task = AzureSQLDBQuery(timeout=self.timeout)
        query_task.bind(
            query=self.query,
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )



import os
from unittest import mock

import pytest

from viadot.tasks import MindfulToCSV


class MockClass:
    status_code = 200
    content = b'[{"id":7277599,"survey_id":505,"phone_number":"","survey_type":"inbound"},{"id":7277294,"survey_id":504,"phone_number":"","survey_type":"web"}]'


@pytest.mark.init
def test_instance_mindful():
    mf = MindfulToCSV()
    assert isinstance(mf, MindfulToCSV)


@mock.patch("viadot.sources.mindful.handle_api_response", return_value=MockClass)
@pytest.mark.run
def test_mindful_run(mock_interactions):
    mf = MindfulToCSV()
    mf.run()
    mock_interactions.call_count == 2
    assert os.path.exists("interactions.csv")
    os.remove("interactions.csv")
    assert os.path.exists("responses.csv")
    os.remove("responses.csv")



import inspect
import json
from datetime import date, timedelta
from typing import List

import pandas as pd
from prefect.utilities import logging

from ..exceptions import CredentialError
from ..utils import handle_api_response
from .base import Source

logger = logging.get_logger(__name__)


class Mediatool(Source):
    """
    Class for downloading data from Mediatool platform. Using Mediatool class user is able to download
    organizations, media entries, campaigns, vehicles, and media types data.
    """

    def __init__(
        self,
        credentials: dict,
        organization_id: str = None,
        user_id: str = None,
        *args,
        **kwargs,
    ):
        """
        Create an instance of the Mediatool class.

        Args:
            credentials (dict): Mediatool credentials. Credentials have to contain authorization 'TOKEN'.
            organization_id (str, optional): Organization ID. Defaults to None.
            user_id (str, optional): User ID. Defaults to None.
        """
        if credentials is not None:
            try:
                self.header = {"Authorization": f"Bearer {credentials.get('TOKEN')}"}
            except:
                raise CredentialError("Credentials not found.")

        super().__init__(*args, credentials=credentials, **kwargs)

        self.organization_id = organization_id or self.credentials.get(
            "ORGANIZATION_ID"
        )
        self.user_id = user_id or self.credentials.get("USER_ID")

    def rename_columns(
        self, df: pd.DataFrame = None, column_suffix: str = "rename"
    ) -> pd.DataFrame:
        """
        Function for renaming columns. Source name is added to the end of the column name to make it unique.

        Args:
            df (pd.DataFrame, optional): Data frame. Defaults to None.
            column_suffix (str, optional): String to be added at the end of column name. Defaults to "rename".

        Returns:
            pd.DataFrame: Final dataframe after changes.
        """
        if isinstance(df, pd.DataFrame):
            column_suffix = column_suffix.split("get_")[-1]
            dict_mapped_names = {
                column_name: f"{column_name}_{column_suffix}"
                for column_name in df.columns
            }
            df_updated = df.rename(columns=dict_mapped_names)
            return df_updated
        else:
            raise TypeError("df object given to the function is not a data frame.")

    def get_media_entries(
        self,
        organization_id: str,
        columns: str = None,
        start_date: str = None,
        end_date: str = None,
        time_delta: int = 360,
        return_dataframe: bool = True,
    ) -> pd.DataFrame:
        """
        Get data for media entries. This is a main function. Media entries contain IDs for most of the fields
        for other endpoints.Returns DataFrame or Dict.

        Args:
            organization_id (str): Organization ID.
            columns (str, optional): What columns should be extracted. Defaults to None.
            start_date (str, optional): Start date e.g '2022-01-01'. Defaults to None.
            end_date (str, optional): End date e.g '2022-01-01'. Defaults to None.
            time_delta (int, optional): The number of days to retrieve from 'today' (today - time_delta). Defaults to 360.
            return_dataframe (bool, optional): Return a dataframe if True. If set to False, return the data as a dict.
                Defaults to True.

        Returns:
            pd.DataFrame: Default return dataframe If 'return_daframe=False' then return list of dicts.
        """
        today = date.today()

        if start_date is None:
            start_date = str(today - timedelta(time_delta))
        if end_date is None:
            start_date = str(today)

        url = f'https://api.mediatool.com/searchmediaentries?q={{"organizationId": "{organization_id}"}}'

        response = handle_api_response(
            url=url,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        if return_dataframe is True:
            df = pd.DataFrame.from_dict(response_dict["mediaEntries"])
            if columns is None:
                columns = df.columns
            try:
                df_filtered = df[columns]
            except KeyError as e:
                logger.info(e)
            return df_filtered

        return response_dict["mediaEntries"]

    def get_campaigns(
        self, organization_id: str, return_dataframe: bool = True
    ) -> pd.DataFrame:
        """
        Get campaign data based on the organization ID. Returns DataFrame or Dict.

        Args:
            organization_id (str): Organization ID.
            return_dataframe (bool, optional): Return a dataframe if True. If set to False, return the data as a dict.
                Defaults to True.

        Returns:
            pd.DataFrame: Default return dataframe If 'return_daframe=False' then return list of dicts.
        """
        url_campaigns = (
            f"https://api.mediatool.com/organizations/{organization_id}/campaigns"
        )

        response = handle_api_response(
            url=url_campaigns,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)

        if return_dataframe is True:
            df = pd.DataFrame.from_dict(response_dict["campaigns"])
            df.replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )
            function_name = inspect.stack()[0][3]
            df_updated = self.rename_columns(df=df, column_suffix=function_name)
            return df_updated

        return response_dict["campaigns"]

    def get_vehicles(
        self,
        vehicle_ids: List[str],
        return_dataframe: bool = True,
    ) -> pd.DataFrame:
        """
        Get vehicles data based on the organization IDs. Returns DataFrame or Dict.

        Args:
            vehicle_ids (List[str]): List of organization IDs.
            return_dataframe (bool, optional): Return a dataframe if True. If set to False, return the data as a dict.
                Defaults to True.

        Returns:
            pd.DataFrame: Default return dataframe. If 'return_daframe=False' then return list of dicts.
        """
        response_dict = {}
        dfs = []
        missing_vehicles = []

        for id in vehicle_ids:
            url = f"https://api.mediatool.com/vehicles/{id}"
            try:
                response = handle_api_response(
                    url=url,
                    headers=self.header,
                    method="GET",
                )
            except Exception:
                missing_vehicles.append(id)

            else:
                response_dict = json.loads(response.text)
                df_single = pd.DataFrame(response_dict["vehicle"], index=[0])
                dfs.append(df_single)

        if missing_vehicles:
            logger.error(f"Vehicle were not found for: {missing_vehicles}.")

        if return_dataframe is True:
            if len(dfs) > 0:
                df = pd.concat(dfs)
                function_name = inspect.stack()[0][3]
                df_updated = self.rename_columns(df=df, column_suffix=function_name)
                return df_updated
            return None

        return response_dict["vehicles"]

    def get_organizations(
        self, user_id: str = None, return_dataframe: bool = True
    ) -> pd.DataFrame:
        """
        Get organizations data based on the user ID. Returns DataFrame or Dict.

        Args:
            user_id (str): User ID.
            return_dataframe (bool, optional): Return a dataframe if True. If set to False, return the data as a dict.
            Defaults to True.

        Returns:
            pd.DataFrame: Default return dataframe. If 'return_daframe=False' then return list of dicts.
        """
        user_id = user_id or self.user_id
        url_organizations = f"https://api.mediatool.com/users/{user_id}/organizations"

        response = handle_api_response(
            url=url_organizations,
            headers=self.header,
            method="GET",
        )
        response_dict = json.loads(response.text)
        organizations = response_dict["organizations"]

        list_organizations = []
        for org in organizations:
            list_organizations.append(
                {
                    "_id": org.get("_id"),
                    "name": org.get("name"),
                    "abbreviation": org.get("abbreviation"),
                }
            )

        if return_dataframe is True:
            df = pd.DataFrame.from_dict(list_organizations)
            function_name = inspect.stack()[0][3]
            df_updated = self.rename_columns(df=df, column_suffix=function_name)
            return df_updated

        return list_organizations

    def get_media_types(
        self, media_type_ids: List[str], return_dataframe: bool = True
    ) -> pd.DataFrame:
        """
        Get media types data based on the media types ID. User have to provide list of media type IDs.
        Returns DataFrame or Dict.

        Args:
            media_type_ids (List[str]): List of media type IDs.
            return_dataframe (bool, optional): Return a dataframe if True. If set to False, return the data as a dict.
                Defaults to True.

        Returns:
            pd.DataFrame: Default return dataframe. If 'return_daframe=False' then return list of dicts.
        """
        list_media_types = []
        for id_media_type in media_type_ids:
            response = handle_api_response(
                url=f"https://api.mediatool.com/mediatypes/{id_media_type}",
                headers=self.header,
                method="GET",
            )
            response_dict = json.loads(response.text)
            list_media_types.append(
                {
                    "_id": response_dict.get("mediaType").get("_id"),
                    "name": response_dict.get("mediaType").get("name"),
                    "type": response_dict.get("mediaType").get("type"),
                }
            )

        if return_dataframe is True:
            df = pd.DataFrame.from_dict(list_media_types)
            function_name = inspect.stack()[0][3]
            df_updated = self.rename_columns(df=df, column_suffix=function_name)
            return df_updated

        return list_media_types



import os

import pytest
from prefect import Flow

from viadot.flows import SQLServerToParquet
from viadot.tasks import SQLServerToDF
from viadot.tasks.sql_server import SQLServerQuery

SCHEMA = "sandbox"
TABLE = "test"
PATH = "test.parquet"


@pytest.fixture(scope="session")
def create_table():
    query_task = SQLServerQuery("AZURE_SQL")
    query_task.run(f"DROP TABLE IF EXISTS {SCHEMA}.{TABLE}")
    query_task.run(f"CREATE TABLE {SCHEMA}.{TABLE} (Id INT, Name VARCHAR (10))")
    yield True


def test_sql_server_to_parquet_flow(create_table):
    flow = SQLServerToParquet(
        name="test_flow",
        sql_query=f"SELECT * FROM {SCHEMA}.{TABLE}",
        local_file_path=PATH,
        if_exists="fail",
        sqlserver_config_key="AZURE_SQL",
        timeout=3600,
    )
    flow.gen_flow()
    assert isinstance(flow, Flow)
    assert len(flow.tasks) == 3  # Number of tasks in the flow
    tasks = list(flow.tasks)

    assert isinstance(tasks[0], SQLServerToDF)
    flow.run()
    assert os.path.isfile(PATH) == True
    os.remove(PATH)



from logging import log
from typing import Any, Dict

import pandas as pd
import prefect
from pendulum import instance
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from ..sources.sqlite import SQLite


class SQLiteInsert(Task):
    """
    Task for inserting data from a pandas DataFrame into SQLite.

    Args:
        db_path (str, optional): The path to the database to be used. Defaults to None.
        sql_path (str, optional): The path to the text file containing the query. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    """

    def __init__(
        self,
        df: pd.DataFrame = None,
        db_path: str = None,
        schema: str = None,
        table_name: str = None,
        if_exists: str = "fail",
        dtypes: Dict[str, Any] = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.db_path = db_path
        self.table_name = table_name
        self.df = df
        self.dtypes = dtypes
        self.schema = schema
        self.if_exists = if_exists

        super().__init__(name="sqlite_insert", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("df", "db_path", "schema", "table_name", "if_exists", "dtypes")
    def run(
        self,
        table_name: str = None,
        schema: str = None,
        dtypes: Dict[str, Any] = None,
        db_path: str = None,
        df: pd.DataFrame = None,
        if_exists: str = "skip",
    ):
        sqlite = SQLite(credentials=dict(db_name=db_path))
        sqlite.create_table(
            table=table_name, schema=schema, dtypes=dtypes, if_exists=if_exists
        )
        logger = prefect.context.get("logger")
        if isinstance(df, pd.DataFrame) == False:
            logger.warning("Object is not a pandas DataFrame")
        elif df.empty:
            logger.warning("DataFrame is empty")
        else:
            sqlite.insert_into(table=table_name, df=df)

        return True


class SQLiteSQLtoDF(Task):
    """
    Task for downloading data from the SQLite to a pandas DataFrame.

    SQLite will create a new database in the directory specified by the 'db_path' parameter.

    Args:
        db_path (str, optional): The path to the database to be used. Defaults to None.
        sql_path (str, optional): The path to the text file containing the query. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    """

    def __init__(
        self,
        db_path: str = None,
        sql_path: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.db_path = db_path
        self.sql_path = sql_path

        super().__init__(name="sqlite_sql_to_df", timeout=timeout, *args, **kwargs)

    def __call__(self):
        """Generate a DataFrame from a SQLite SQL query"""

    @defaults_from_attrs("db_path", "sql_path")
    def run(self, sql_path: str = None, db_path: str = None) -> pd.DataFrame:
        sqlite = SQLite(credentials=dict(db_name=db_path))

        logger = prefect.context.get("logger")
        logger.info(f"Loading a DataFrame from {sql_path}...")

        with open(sql_path, "r") as queryfile:
            query = queryfile.read()

        df = sqlite.to_df(query)
        logger.info(f"DataFrame was successfully loaded.")

        return df


class SQLiteQuery(Task):
    """
    Task for running an SQLite query.

    Args:
        query (str, optional): The query to execute on the database. Defaults to None.
        db_path (str, optional): The path to the database to be used. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    """

    def __init__(
        self,
        query: str = None,
        db_path: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.query = query
        self.db_path = db_path
        super().__init__(name="sqlite_query", timeout=timeout, *args, **kwargs)

    def __call__(self):
        """Run an SQL query on SQLite"""

    @defaults_from_attrs("query", "db_path")
    def run(self, query: str = None, db_path: str = None):
        sqlite = SQLite(credentials=dict(db_name=db_path))

        # run the query and fetch the results if it's a select
        result = sqlite.run(query=query)

        self.logger.info(f"Successfully ran the query.")

        return result



import re
from typing import Any, List, Literal, NoReturn, Tuple, Union

import duckdb
import pandas as pd
from prefect.utilities import logging

from ..config import local_config
from ..exceptions import CredentialError
from ..signals import SKIP
from .base import Source

logger = logging.get_logger(__name__)

Record = Tuple[Any]


class DuckDB(Source):
    DEFAULT_SCHEMA = "main"

    def __init__(
        self,
        config_key: str = "DuckDB",
        credentials: dict = None,
        *args,
        **kwargs,
    ):
        """A class for interacting with DuckDB.

        Args:
            config_key (str, optional): The key inside local config containing the config.
            User can choose to use this or pass credentials directly to the `credentials`
            parameter. Defaults to None.
            credentials (dict, optional): Credentials for the connection. Defaults to None.
        """

        if config_key:
            config_credentials = local_config.get(config_key)

        credentials = credentials if credentials else config_credentials
        if credentials is None:
            raise CredentialError("Credentials not found.")

        super().__init__(*args, credentials=credentials, **kwargs)

    @property
    def con(self) -> duckdb.DuckDBPyConnection:
        """Return a new connection to the database. As the views are highly isolated,
        we need a new connection for each query in order to see the changes from
        previous queries (eg. if we create a new table and then we want to list
        tables from INFORMATION_SCHEMA, we need to create a new DuckDB connection).

        Returns:
            duckdb.DuckDBPyConnection: database connection.
        """
        return duckdb.connect(
            database=self.credentials.get("database"),
            read_only=self.credentials.get("read_only", False),
        )

    @property
    def tables(self) -> List[str]:
        """Show the list of fully qualified table names.

        Returns:
            List[str]: The list of tables in the format '{SCHEMA}.{TABLE}'.
        """
        tables_meta: List[Tuple] = self.run("SELECT * FROM information_schema.tables")
        tables = [table_meta[1] + "." + table_meta[2] for table_meta in tables_meta]
        return tables

    @property
    def schemas(self) -> List[str]:
        """Show the list of schemas.

        Returns:
            List[str]: The list ofschemas.
        """
        self.logger.warning(
            "DuckDB does not expose a way to list schemas. `DuckDB.schemas` only contains schemas with tables."
        )
        tables_meta: List[Tuple] = self.run("SELECT * FROM information_schema.tables")
        schemas = [table_meta[1] for table_meta in tables_meta]
        return schemas

    def to_df(self, query: str, if_empty: str = None) -> pd.DataFrame:
        if query.upper().startswith("SELECT"):
            df = self.run(query, fetch_type="dataframe")
            if df.empty:
                self._handle_if_empty(if_empty=if_empty)
        else:
            df = pd.DataFrame()
        return df

    def run(
        self, query: str, fetch_type: Literal["record", "dataframe"] = "record"
    ) -> Union[List[Record], bool]:
        """Run a query on DuckDB.

        Args:
            query (str): The query to execute.
            fetch_type (Literal[, optional): How to return the data: either
            in the default record format or as a pandas DataFrame. Defaults to "record".

        Returns:
            Union[List[Record], bool]: Either the result set of a query or,
            in case of DDL/DML queries, a boolean describing whether
            the query was excuted successfuly.
        """
        allowed_fetch_type_values = ["record", "dataframe"]
        if fetch_type not in allowed_fetch_type_values:
            raise ValueError(
                f"Only the values {allowed_fetch_type_values} are allowed for 'fetch_type'"
            )
        cursor = self.con.cursor()
        cursor.execute(query)

        query_clean = query.upper().strip()
        regex = r"^\s*[--;].*"
        lines = query_clean.splitlines()
        final_query = ""

        for line in lines:
            line = line.strip()
            match_object = re.match(regex, line)
            if not match_object:
                final_query += " " + line
        final_query = final_query.strip()
        query_keywords = ["SELECT", "SHOW", "PRAGMA", "WITH"]
        if any(final_query.startswith(word) for word in query_keywords):
            if fetch_type == "record":
                result = cursor.fetchall()
            else:
                result = cursor.fetchdf()
        else:
            result = True

        cursor.close()
        return result

    def _handle_if_empty(self, if_empty: str = "warn") -> NoReturn:
        if if_empty == "warn":
            logger.warning("The query produced no data.")
        elif if_empty == "skip":
            raise SKIP("The query produced no data. Skipping...")
        elif if_empty == "fail":
            raise ValueError("The query produced no data.")

    def create_table_from_parquet(
        self,
        table: str,
        path: str,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
    ) -> NoReturn:
        """Create a DuckDB table with a CTAS from Parquet file(s).

        Args:
            table (str): Destination table.
            path (str): The path to the source Parquet file(s). Glob expressions are
            also allowed here (eg. `my_folder/*.parquet`).
            schema (str, optional): Destination schema. Defaults to None.
            if_exists (Literal[, optional): What to do if the table already exists. Defaults to "fail".

        Raises:
            ValueError: If the table exists and `if_exists` is set to `fail`.

        Returns:
            NoReturn: Does not return anything.
        """
        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table
        exists = self._check_if_table_exists(schema=schema, table=table)

        if exists:
            if if_exists == "replace":
                self.run(f"DROP TABLE {fqn}")
            elif if_exists == "append":
                self.logger.info(f"Appending to table {fqn}...")
                ingest_query = f"COPY {fqn} FROM '{path}' (FORMAT 'parquet')"
                self.run(ingest_query)
                self.logger.info(f"Successfully appended data to table '{fqn}'.")
                return True
            elif if_exists == "delete":
                self.run(f"DELETE FROM {fqn}")
                self.logger.info(f"Successfully deleted data from table '{fqn}'.")
                self.run(f"INSERT INTO {fqn} SELECT * FROM read_parquet('{path}')")
                self.logger.info(f"Successfully inserted data into table '{fqn}'.")
                return True
            elif if_exists == "fail":
                raise ValueError(
                    "The table already exists and 'if_exists' is set to 'fail'."
                )
            elif if_exists == "skip":
                return False

        self.run(f"CREATE SCHEMA IF NOT EXISTS {schema}")

        self.logger.info(f"Creating table {fqn}...")
        ingest_query = f"CREATE TABLE {fqn} AS SELECT * FROM '{path}';"
        self.run(ingest_query)
        self.logger.info(f"Table {fqn} has been created successfully.")

    def drop_table(self, table: str, schema: str = None) -> bool:
        """
        Drop a table.

        This is a thin wraper around DuckDB.run() which logs to the operation.

        Args:
            table (str): The table to be dropped.
            schema (str, optional): The schema where the table is located.
            Defaults to None.

        Returns:
            bool: Whether the table was dropped.
        """

        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table

        self.logger.info(f"Dropping table {fqn}...")
        dropped = self.run(f"DROP TABLE IF EXISTS {fqn}")
        if dropped:
            self.logger.info(f"Table {fqn} has been dropped successfully.")
        else:
            self.logger.info(f"Table {fqn} could not be dropped.")
        return dropped

    def _check_if_table_exists(self, table: str, schema: str = None) -> bool:
        schema = schema or DuckDB.DEFAULT_SCHEMA
        fqn = schema + "." + table
        return fqn in self.tables

    def _check_if_schema_exists(self, schema: str) -> bool:
        if schema == self.DEFAULT_SCHEMA:
            return True
        fqns = self.tables
        return any((fqn.split(".")[0] == schema for fqn in fqns))



import json
import logging
import os

import pytest
from prefect.tasks.secrets import PrefectSecret

from viadot.flows import DuckDBToSQLServer
from viadot.sources import DuckDB
from viadot.tasks.azure_key_vault import AzureKeyVaultSecret

TABLE = "test_table"
DUCKDB_SCHEMA = "test_schema"
SQL_SERVER_SCHEMA = "sandbox"
TABLE_MULTIPLE_PARQUETS = "test_multiple_parquets"
DATABASE_PATH = "test_db_123.duckdb"


@pytest.fixture(scope="session")
def duckdb():
    duckdb = DuckDB(credentials=dict(database=DATABASE_PATH))
    yield duckdb
    os.remove(DATABASE_PATH)


def test_create_table_from_parquet(duckdb, TEST_PARQUET_FILE_PATH, caplog):
    with caplog.at_level(logging.INFO):
        duckdb.create_table_from_parquet(
            schema=DUCKDB_SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
        )

    assert "created successfully" in caplog.text


def test_duckdb_sql_server_flow():
    credentials_secret = PrefectSecret(
        "AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET"
    ).run()
    vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()
    azure_secret_task = AzureKeyVaultSecret()
    credentials_str = azure_secret_task.run(
        secret=credentials_secret, vault_name=vault_name
    )

    flow = DuckDBToSQLServer(
        "test_duckdb_flow_run",
        duckdb_credentials=dict(database=DATABASE_PATH),
        sql_server_credentials=json.loads(credentials_str),
        sql_server_schema="sandbox",
        sql_server_table=TABLE,
        duckdb_schema=DUCKDB_SCHEMA,
        duckdb_table=TABLE,
    )
    r = flow.run()
    assert r.is_successful()



import json
from datetime import datetime
from typing import Literal

import pandas as pd
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.sources.customer_gauge import CustomerGauge
from viadot.tasks.azure_key_vault import AzureKeyVaultSecret

logger = logging.get_logger()


class CustomerGaugeToDF(Task):
    def __init__(
        self,
        endpoint: Literal["responses", "non-responses"] = None,
        total_load: bool = True,
        endpoint_url: str = None,
        cursor: int = None,
        pagesize: int = 1000,
        date_field: Literal[
            "date_creation", "date_order", "date_sent", "date_survey_response"
        ] = None,
        start_date: datetime = None,
        end_date: datetime = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Task CustomerGaugeToDF for downloading the selected range of data from Customer Gauge endpoint and return as one pandas DataFrame.

        Args:
            endpoint (Literal["responses", "non-responses"], optional): Indicate which endpoint to connect. Defaults to None.
            total_load (bool, optional): Indicate whether to download the data to the latest. If 'False', only one API call is executed (up to 1000 records). Defaults to True.
            endpoint_url (str, optional): Endpoint URL. Defaults to None.
            cursor (int, optional): Cursor value to navigate to the page. Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page, max value = 1000. Defaults to 1000.
            date_field (Literal["date_creation", "date_order", "date_sent", "date_survey_response"], optional): Specifies the date type which filter date range. Defaults to None.
            start_date (datetime, optional): Defines the period end date in yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period start date in yyyy-mm-dd format. Defaults to None.
            timeout (int, optional): The time (in seconds) to wait while running this task before a timeout occurs. Defaults to 3600.
        """
        self.endpoint = endpoint
        self.total_load = total_load
        self.endpoint_url = endpoint_url
        self.cursor = cursor
        self.pagesize = pagesize
        self.date_field = date_field
        self.start_date = start_date
        self.end_date = end_date

        super().__init__(
            name="customer_gauge_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download Customer Gauge data to a DF"""
        super().__call__(self)

    @defaults_from_attrs(
        "endpoint",
        "total_load",
        "endpoint_url",
        "cursor",
        "pagesize",
        "date_field",
        "start_date",
        "end_date",
    )
    def run(
        self,
        endpoint: Literal["responses", "non-responses"] = None,
        total_load: bool = True,
        endpoint_url: str = None,
        cursor: int = None,
        pagesize: int = 1000,
        date_field: Literal[
            "date_creation", "date_order", "date_sent", "date_survey_response"
        ] = None,
        start_date: datetime = None,
        end_date: datetime = None,
        credentials_secret: str = "CUSTOMER-GAUGE",
        vault_name: str = None,
    ) -> pd.DataFrame:
        """
        Run method. Downloading the selected range of data from Customer Gauge endpoint and return as one pandas DataFrame.

        Args:
            endpoint (Literal["responses", "non-responses"]): Indicate which endpoint to connect. Defaults to None.
            total_load (bool, optional): Indicate whether to download the data to the latest. If 'False', only one API call is executed (up to 1000 records). Defaults to True.
            endpoint_url (str, optional): Endpoint URL. Defaults to None.
            cursor (int, optional): Cursor value to navigate to the page. Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page, max value = 1000. Defaults to 1000.
            date_field (Literal["date_creation", "date_order", "date_sent", "date_survey_response"], optional): Specifies the date type which filter date range. Defaults to None.
            start_date (datetime, optional): Defines the period end date in yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period start date in yyyy-mm-dd format. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with ['client_id', 'client_secret']. Defaults to "CUSTOMER-GAUGE".
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

        Returns:
            pd.DataFrame: Final pandas DataFrame.
        """
        try:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)
        except (ValueError, TypeError) as e:
            logger.error(e)

        df_list = []

        customer_gauge = CustomerGauge(
            endpoint=endpoint, url=endpoint_url, credentials=credentials
        )
        logger.info(
            f"Starting downloading data from {self.endpoint or self.endpoint_url} endpoint..."
        )
        json_data = customer_gauge.get_json_response(
            cursor=cursor,
            pagesize=pagesize,
            date_field=date_field,
            start_date=start_date,
            end_date=end_date,
        )
        cur = customer_gauge.get_cursor(json_data)
        df = customer_gauge.to_df(json_data)
        df_list.append(df)
        if total_load == True:
            if cursor is None:
                logger.info(
                    f"Downloading all the data from the {self.endpoint or self.endpoint_url} endpoint. Process might take a few minutes..."
                )
            else:
                logger.info(
                    f"Downloading starting from the {cursor} cursor. Process might take a few minutes..."
                )
            while df.empty == False:
                json_data = customer_gauge.get_json_response(cursor=cur)
                cur = customer_gauge.get_cursor(json_data)
                df = customer_gauge.to_df(json_data)
                df_list.append(df)

        df_total = pd.concat(df_list, ignore_index=True)

        return df_total



from typing import Any, List, Literal, NoReturn, Tuple, Union

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from ..signals import SKIP
from ..sources import DuckDB
from ..utils import check_if_empty_file

Record = Tuple[Any]


class DuckDBQuery(Task):
    """
    Task for running a query on DuckDB.

    Args:
        credentials (dict, optional): The config to use for connecting with the db.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        credentials: dict = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.credentials = credentials
        super().__init__(name="run_duckdb_query", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("credentials")
    def run(
        self,
        query: str,
        fetch_type: Literal["record", "dataframe"] = "record",
        credentials: dict = None,
    ) -> Union[List[Record], bool]:
        """Run a query on DuckDB.

        Args:
            query (str, required):  The query to execute.
            fetch_type (Literal[, optional): How to return the data: either
            in the default record format or as a pandas DataFrame. Defaults to "record".
            credentials (dict, optional): The config to use for connecting with the db.

        Returns:
            Union[List[Record], bool]: Either the result set of a query or,
            in case of DDL/DML queries, a boolean describing whether
            the query was excuted successfuly.
        """

        duckdb = DuckDB(credentials=credentials)

        # run the query and fetch the results if it's a select
        result = duckdb.run(query, fetch_type=fetch_type)

        self.logger.info(f"Successfully ran the query.")
        return result


class DuckDBCreateTableFromParquet(Task):
    """
    Task for creating a DuckDB table with a CTAS from Parquet file(s).

    Args:
        table (str, optional): Destination table.
        also allowed here (eg. `my_folder/*.parquet`).
        schema (str, optional): Destination schema.
        if_exists (Literal, optional): What to do if the table already exists.
        if_empty (Literal, optional): What to do if ".parquet" file is emty. Defaults to "skip".
        credentials(dict, optional): The config to use for connecting with the db.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    Raises:
        ValueError: If the table exists and `if_exists`is set to `fail` or when parquet file
        is empty and `if_empty` is set to `fail`.

    Returns:
        NoReturn: Does not return anything.
    """

    def __init__(
        self,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
        if_empty: Literal["skip", "fail"] = "skip",
        credentials: dict = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.schema = schema
        self.if_exists = if_exists
        self.if_empty = if_empty
        self.credentials = credentials

        super().__init__(
            name="duckdb_create_table",
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs("schema", "if_exists", "if_empty")
    def run(
        self,
        table: str,
        path: str,
        schema: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = None,
        if_empty: Literal["skip", "fail"] = None,
    ) -> NoReturn:
        """
        Create a DuckDB table with a CTAS from Parquet file(s).

        Args:
            table (str, optional): Destination table.
            path (str): The path to the source Parquet file(s). Glob expressions are
            also allowed here (eg. `my_folder/*.parquet`).
            schema (str, optional): Destination schema.
            if_exists (Literal, optional): What to do if the table already exists.
            if_empty (Literal, optional): What to do if Parquet file is empty. Defaults to None.

        Raises:
            ValueError: If the table exists and `if_exists`is set to `fail` or when parquet file
            is empty and `if_empty` is set to `fail`.

        Returns:
            NoReturn: Does not return anything.
        """
        try:
            check_if_empty_file(path=path, if_empty=if_empty)
        except SKIP:
            self.logger.info("The input file is empty. Skipping.")
            return

        duckdb = DuckDB(credentials=self.credentials)

        fqn = f"{schema}.{table}" if schema is not None else table
        created = duckdb.create_table_from_parquet(
            path=path, schema=schema, table=table, if_exists=if_exists
        )
        if created:
            self.logger.info(f"Successfully created table {fqn}.")
        else:
            self.logger.info(
                f"Table {fqn} has not been created as if_exists is set to {if_exists}."
            )


class DuckDBToDF(Task):
    """
    Load a table from DuckDB to a pandas DataFrame.

    Args:
        schema (str, optional): Source schema.
        table (str, optional): Source table.
        if_empty (Literal[, optional): What to do if the query returns no data.
        Defaults to "warn".
        credentials (dict, optional): The config to use for connecting with the db.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    Returns:
        pd.DataFrame: a pandas DataFrame containing the table data.
    """

    def __init__(
        self,
        schema: str = None,
        table: str = None,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        credentials: dict = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.schema = schema
        self.table = table
        self.if_empty = if_empty
        self.credentials = credentials

        super().__init__(name="duckdb_to_df", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("schema", "table", "if_empty", "credentials")
    def run(
        self,
        schema: str = None,
        table: str = None,
        if_empty: Literal["warn", "skip", "fail"] = None,
        credentials: dict = None,
    ) -> pd.DataFrame:
        """Load a DuckDB table into a pandas DataFrame.

        Args:
            schema (str, optional): Source schema.
            table (str, optional): Source table.
            if_empty (Literal[, optional): What to do if the query returns no data.
            Defaults to "warn".
            credentials (dict, optional): The config to use for connecting with the db.

        Returns:
            pd.DataFrame: a pandas DataFrame containing the table data.
        """

        if table is None:
            raise ValueError("Table is required.")

        duckdb = DuckDB(credentials=credentials)

        # run the query and fetch the results if it's a select
        fqn = f"{schema}.{table}" if schema is not None else table
        query = f"SELECT * FROM {fqn}"
        df = duckdb.to_df(query, if_empty=if_empty)

        self.logger.info(f"Data has been loaded sucessfully.")
        return df



import os
from typing import Any, Dict, List, Union

import pendulum
from prefect import Flow, Task, apply_map

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_to_csv,
    df_to_parquet,
    union_dfs_task,
)
from viadot.tasks import AzureDataLakeUpload, C4CReportToDF, C4CToDF
from viadot.utils import slugify


class CloudForCustomersReportToADLS(Flow):
    def __init__(
        self,
        name: str = None,
        report_url: str = None,
        url: str = None,
        endpoint: str = None,
        params: Dict[str, Any] = {},
        fields: List[str] = None,
        skip: int = 0,
        top: int = 1000,
        channels: List[str] = None,
        months: List[str] = None,
        years: List[str] = None,
        env: str = "QA",
        c4c_credentials_secret: str = None,
        local_file_path: str = None,
        output_file_extension: str = ".csv",
        adls_dir_path: str = None,
        adls_file_path: str = None,
        overwrite_adls: bool = False,
        adls_sp_credentials_secret: str = None,
        if_empty: str = "warn",
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from different marketing APIs to a local CSV
        using Cloud for Customers API, then uploading it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            report_url (str, optional): The url to the API. Defaults to None.
            url (str, optional): ???
            endpoint (str, optional): ???
            params (dict, optional): ???
            fields (list, optional): ???
            skip (int, optional): Initial index value of reading row. Defaults to 0.
            top (int, optional): The value of top reading row. Defaults to 1000.
            channels (List[str], optional): Filtering parameters passed to the url. Defaults to None.
            months (List[str], optional): Filtering parameters passed to the url. Defaults to None.
            years (List[str], optional): Filtering parameters passed to the url. Defaults to None.
            env (str, optional): ???
            c4c_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            username and password for the Cloud for Customers instance.
            local_file_path (str, optional): Local destination path. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data which is not easy
            to handle with parquet. Defaults to ".csv".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            adls_file_path (str, optional): Azure Data Lake destination file path. Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to False.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            if_empty (str, optional): What to do if the Supermetrics query returns no data. Defaults to "warn".
            if_exists (str, optional): What to do if the local file already exists. Defaults to "replace".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.report_url = report_url
        self.skip = skip
        self.top = top
        self.if_empty = if_empty
        self.env = env
        self.c4c_credentials_secret = c4c_credentials_secret
        self.timeout = timeout

        # AzureDataLakeUpload
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.overwrite_adls = overwrite_adls
        self.output_file_extension = output_file_extension
        self.local_file_path = (
            local_file_path or slugify(name) + self.output_file_extension
        )
        self.now = str(pendulum.now("utc"))
        self.adls_dir_path = adls_dir_path
        self.adls_file_path = adls_file_path or os.path.join(
            adls_dir_path, self.now + self.output_file_extension
        )

        # in case of non-report invoking
        self.url = url
        self.endpoint = endpoint
        self.params = params
        self.fields = fields

        # filtering report_url for reports
        self.channels = channels
        self.months = months
        self.years = years

        self.report_urls_with_filters = [self.report_url]

        self.report_urls_with_filters = self.create_url_with_fields(
            fields_list=self.channels, filter_code="CCHANNETZTEXT12CE6C2FA0D77995"
        )

        self.report_urls_with_filters = self.create_url_with_fields(
            fields_list=self.months, filter_code="CMONTH_ID"
        )

        self.report_urls_with_filters = self.create_url_with_fields(
            fields_list=self.years, filter_code="CYEAR_ID"
        )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def create_url_with_fields(self, fields_list: List[str], filter_code: str) -> List:
        urls_list_result = []
        add_filter = True
        if len(self.report_urls_with_filters) > 1:
            add_filter = False

        if fields_list:
            for url in self.report_urls_with_filters:
                for field in fields_list:
                    if add_filter:
                        new_url = f"{url}&$filter=({filter_code}%20eq%20%27{field}%27)"
                    elif not add_filter:
                        new_url = f"{url}%20and%20({filter_code}%20eq%20%27{field}%27)"
                    urls_list_result.append(new_url)
            return urls_list_result
        else:
            return self.report_urls_with_filters

    def gen_c4c(
        self,
        url: str,
        endpoint: str,
        params: str,
        env: str,
        credentials_secret: str = None,
        flow: Flow = None,
    ) -> Task:
        c4c_to_df = C4CToDF(timeout=self.timeout)
        df = c4c_to_df.bind(
            url=url,
            endpoint=endpoint,
            params=params,
            env=env,
            credentials_secret=credentials_secret,
            flow=flow,
        )

        return df

    def gen_c4c_report_months(
        self, report_urls_with_filters: Union[str, List[str]], flow: Flow = None
    ) -> Task:
        c4c_report_to_df = C4CReportToDF(timeout=self.timeout)
        report = c4c_report_to_df.bind(
            report_url=report_urls_with_filters,
            skip=self.skip,
            top=self.top,
            env=self.env,
            credentials_secret=self.c4c_credentials_secret,
            flow=flow,
        )

        return report

    def gen_flow(self) -> Flow:
        if self.report_url:
            dfs = apply_map(
                self.gen_c4c_report_months, self.report_urls_with_filters, flow=self
            )
            df = union_dfs_task.bind(dfs, flow=self)
        elif self.url:
            df = self.gen_c4c(
                url=self.url,
                endpoint=self.endpoint,
                params=self.params,
                env=self.env,
                credentials_secret=self.c4c_credentials_secret,
                flow=self,
            )

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_with_metadata.set_upstream(df, flow=self)
        df_to_file.set_upstream(df_with_metadata, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)



import json
import time
from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal

from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.config import local_config
from viadot.exceptions import CredentialError
from viadot.sources import Mindful
from viadot.tasks import AzureKeyVaultSecret

logger = logging.get_logger()


class MindfulToCSV(Task):
    def __init__(
        self,
        report_name: str = "mindful_to_csv",
        start_date: datetime = None,
        end_date: datetime = None,
        date_interval: int = 1,
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        file_extension: Literal["parquet", "csv"] = "csv",
        file_path: str = "",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """Task for downloading data from Mindful API to CSV.

        Args:
            report_name (str, optional): The name of this task. Defaults to "mindful_to_csv".
            start_date (datetime, optional): Start date of the request. Defaults to None.
            end_date (datetime, optional): End date of the resquest. Defaults to None.
            date_interval (int, optional): How many days are included in the request.
                If end_date is passed as an argument, date_interval will be invalidated. Defaults to 1.
            region (Literal[us1, us2, us3, ca1, eu1, au1], optional): SD region from where to interact with the mindful API. Defaults to "eu1".
            file_extension (Literal[parquet, csv], optional): File extensions for storing responses. Defaults to "csv".
            file_path (str, optional): Path where to save the file locally. Defaults to ''.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

        Raises:
            CredentialError: If credentials are not provided in local_config or directly as a parameter inside run method.
        """

        self.start_date = start_date
        self.end_date = end_date
        self.date_interval = date_interval
        self.region = region
        self.file_extension = file_extension
        self.file_path = file_path

        super().__init__(
            name=report_name,
            timeout=timeout,
            *args,
            **kwargs,
        )

        if not isinstance(start_date, datetime):
            self.start_date = datetime.now() - timedelta(days=date_interval)
            self.end_date = self.start_date + timedelta(days=date_interval)
        elif isinstance(start_date, datetime) and not isinstance(end_date, datetime):
            self.start_date = start_date
            self.end_date = start_date + timedelta(days=date_interval)
            if self.end_date > datetime.now():
                self.end_date = datetime.now()
        elif start_date >= end_date:
            raise ValueError(
                f"start_date variable must be lower than end_date variable."
            )
        else:
            self.start_date = start_date
            self.end_date = end_date

    def __call__(self, *args, **kwargs):
        """Download Mindful data to CSV"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "start_date",
        "end_date",
        "date_interval",
        "region",
        "file_extension",
        "file_path",
    )
    def run(
        self,
        credentials_mindful: Dict[str, Any] = None,
        credentials_secret: str = None,
        vault_name: str = None,
        start_date: datetime = None,
        end_date: datetime = None,
        date_interval: int = 1,
        file_extension: Literal["parquet", "csv"] = "csv",
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        file_path: str = "",
    ):
        if credentials_mindful is not None:
            self.logger.info("Mindful credentials provided by user")
        elif credentials_mindful is None and credentials_secret is not None:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials_mindful = json.loads(credentials_str)
            logger.info("Loaded credentials from Key Vault.")
        else:
            try:
                credentials_mindful = local_config["MINDFUL"]
                self.logger.info("Mindful credentials loaded from local config")
            except KeyError:
                credentials_mindful = None
                raise CredentialError("Credentials not found.")

        auth = (
            credentials_mindful["CUSTOMER_UUID"],
            credentials_mindful["AUTH_TOKEN"],
        )

        mindful = Mindful(
            auth=auth,
            region=region,
            start_date=start_date,
            end_date=end_date,
            date_interval=date_interval,
            file_extension=file_extension,
        )

        file_names = []
        # interactions
        interactions_response = mindful.get_interactions_list()
        if interactions_response.status_code == 200:
            interaction_file_name = mindful.response_to_file(
                interactions_response,
                file_path=file_path,
            )
            file_names.append(interaction_file_name)
            logger.info(
                "Successfully downloaded interactions data from the Mindful API."
            )
            time.sleep(0.5)

        # responses
        responses_response = mindful.get_responses_list()
        if responses_response.status_code == 200:
            response_file_name = mindful.response_to_file(
                responses_response,
                file_path=file_path,
            )
            file_names.append(response_file_name)
            logger.info("Successfully downloaded responses data from the Mindful API.")
            time.sleep(0.5)

        # surveys
        surveys_response = mindful.get_survey_list()
        if surveys_response.status_code == 200:
            surveys_file_name = mindful.response_to_file(
                surveys_response,
                file_path=file_path,
            )
            file_names.append(surveys_file_name)
            logger.info("Successfully downloaded surveys data from the Mindful API.")

        if not file_names:
            return None
        else:
            return file_names



import datetime
import os

import pyarrow
import pytest

from viadot.sources.uk_carbon_intensity import UKCarbonIntensity

LOCAL_TESTS_PATH = "/home/viadot/tests"
TEST_FILE_1 = os.path.join(LOCAL_TESTS_PATH, "tests_out.csv")
TEST_FILE_2 = os.path.join(LOCAL_TESTS_PATH, "testfile_stats.csv")


@pytest.fixture(scope="session")
def carbon():
    carbon = UKCarbonIntensity()
    yield carbon
    os.remove(TEST_FILE_1)
    os.remove(TEST_FILE_2)


def test_to_json(carbon):
    carbon.query("/intensity")
    data = carbon.to_json()
    assert "data" in data


def test_to_df(carbon):
    carbon.query("/intensity")
    df = carbon.to_df()
    assert not df.empty


def test_to_arrow(carbon):
    carbon.query("/intensity")
    table = carbon.to_arrow()
    assert type(table) == pyarrow.lib.Table


def test_to_csv(carbon):
    carbon.query("/intensity")
    carbon.to_csv(TEST_FILE_1)
    assert os.path.isfile(TEST_FILE_1) == True


def test_stats_to_csv(carbon):
    now = datetime.datetime.now()
    for i in range(3):
        from_delta = datetime.timedelta(days=i + 1)
        to_delta = datetime.timedelta(days=i)
        to = now - to_delta
        from_ = now - from_delta
        carbon.query(f"/intensity/stats/{from_.isoformat()}/{to.isoformat()}")
        carbon.to_csv(TEST_FILE_2, if_exists="append")
    assert os.path.isfile(TEST_FILE_2) == True


def test_to_df_today(carbon):
    carbon.query("/intensity/date")
    df = carbon.to_df()
    assert len(df) > 1



from viadot.config import local_config
from viadot.flows import SAPRFCToADLS
from viadot.sources import AzureDataLake
from viadot.tasks import AzureDataLakeRemove

try:
    import pyrfc
except ModuleNotFoundError:
    raise

ADLS_PATH = "raw/supermetrics/mp/test_file_sap.parquet"
FILE_NAME = "test_file.parquet"


def test_sap_rfc_to_adls_query():
    sap_test_creds = local_config.get("SAP").get("QA")
    flow = SAPRFCToADLS(
        name="test flow",
        query="SELECT MATNR, MATKL FROM MARA WHERE LAEDA LIKE '2022%'",
        func="BBP_RFC_READ_TABLE",
        sap_credentials=sap_test_creds,
        local_file_path=FILE_NAME,
        adls_path=ADLS_PATH,
        overwrite=True,
    )
    result = flow.run()
    assert result.is_successful()
    file = AzureDataLake(ADLS_PATH)
    assert file.exists()
    AzureDataLakeRemove(ADLS_PATH)



from viadot.tasks import CloneRepo


def test_create_clone_repo():
    clone = CloneRepo(
        url="https://github.com/organization_name/repo_name.git",
        token="abc123",
    )
    assert isinstance(clone, CloneRepo)
    assert clone.token == "abc123"



from typing import Any, Dict, List, Literal

from prefect import Flow

from viadot.task_utils import add_ingestion_metadata_task, cast_df_to_str, df_to_parquet
from viadot.tasks import DuckDBCreateTableFromParquet, EpicorOrdersToDF


class EpicorOrdersToDuckDB(Flow):
    def __init__(
        self,
        name: str,
        base_url: str,
        filters_xml: str,
        local_file_path: str,
        epicor_credentials: Dict[str, Any] = None,
        epicor_config_key: str = None,
        validate_date_filter: bool = True,
        start_date_field: str = "BegInvoiceDate",
        end_date_field: str = "EndInvoiceDate",
        duckdb_table: str = None,
        duckdb_schema: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
        if_empty: Literal["warn", "skip", "fail"] = "skip",
        duckdb_credentials: dict = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading orders data from Epicor API and uploading it to DuckDB using Parquet files.

        Args:
            base_url (str, required): Base url to Epicor Orders.
            filters_xml (str, required): Filters in form of XML. The date filter is required.
            local_file_path (str): The path to the source Parquet file.
            epicor_credentials (Dict[str, Any], optional): Credentials to connect with Epicor API containing host, port,
                username and password. Defaults to None.
            epicor_config_key (str, optional): Credential key to dictionary where details are stored. Defaults to None.
            validate_date_filter (bool, optional): Whether or not validate xml date filters. Defaults to True.
            start_date_field (str, optional) The name of filters field containing start date. Defaults to "BegInvoiceDate".
            end_date_field (str, optional) The name of filters field containing end date. Defaults to "EndInvoiceDate".
            duckdb_table (str, optional): Destination table in DuckDB. Defaults to None.
            duckdb_schema (str, optional): Destination schema in DuckDB. Defaults to None.
            if_exists (Literal, optional):  What to do if the table already exists. Defaults to "fail".
            if_empty (Literal, optional): What to do if Parquet file is empty. Defaults to "skip".
            duckdb_credentials (dict, optional): Credentials for the DuckDB connection. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.base_url = base_url
        self.epicor_credentials = epicor_credentials
        self.epicor_config_key = epicor_config_key
        self.validate_date_filter = validate_date_filter
        self.filters_xml = filters_xml
        self.end_date_field = end_date_field
        self.start_date_field = start_date_field
        self.local_file_path = local_file_path
        self.duckdb_table = duckdb_table
        self.duckdb_schema = duckdb_schema
        self.if_exists = if_exists
        self.if_empty = if_empty
        self.duckdb_credentials = duckdb_credentials

        super().__init__(*args, name=name, **kwargs)

        self.df_task = EpicorOrdersToDF(
            base_url=self.base_url,
            filters_xml=self.filters_xml,
            timeout=timeout,
        )
        self.create_duckdb_table_task = DuckDBCreateTableFromParquet(
            credentials=duckdb_credentials,
            timeout=timeout,
        )

        self.gen_flow()

    def gen_flow(self) -> Flow:
        df = self.df_task.bind(
            flow=self,
            credentials=self.epicor_credentials,
            config_key=self.epicor_config_key,
            validate_date_filter=self.validate_date_filter,
            end_date_field=self.end_date_field,
            start_date_field=self.start_date_field,
        )
        df_mapped = cast_df_to_str.bind(df, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df_mapped, flow=self)
        parquet = df_to_parquet.bind(
            df=df_with_metadata,
            path=self.local_file_path,
            if_exists=self.if_exists,
            flow=self,
        )
        create_duckdb_table = self.create_duckdb_table_task.bind(
            path=self.local_file_path,
            schema=self.duckdb_schema,
            table=self.duckdb_table,
            if_exists=self.if_exists,
            if_empty=self.if_empty,
            flow=self,
        )
        create_duckdb_table.set_upstream(parquet, flow=self)



import os
from pathlib import Path
from typing import Any, Dict, List, Literal

import pendulum
from prefect import Flow
from prefect.backend import set_key_value

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, SAPBWToDF


class SAPBWToADLS(Flow):
    def __init__(
        self,
        name: str,
        mdx_query: str,
        mapping_dict: dict = None,
        sapbw_credentials: dict = None,
        sapbw_credentials_key: str = "SAP",
        env: str = "BW",
        output_file_extension: str = ".parquet",
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_dir_path: str = None,
        if_exists: Literal["replace", "append", "delete"] = "replace",
        overwrite_adls: bool = True,
        vault_name: str = None,
        sp_credentials_secret: str = None,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from SAP BW to file, then uploading it to ADLS.

        Args:
            name (str): Name of the flow.
            mdx_query (str): MDX query to be passed to SAP BW server.
            mapping_dict (dict, optional): Dictionary with original column names and the mapping for them. If not None then flows is generating mapping automatically with mapping applied by user, if not - it generates automatically the json file with columns.
            sapbw_credentials (dict, optional): Credentials to SAP in dictionary format. Defaults to None.
            sapbw_credentials_key (str, optional): Azure KV secret. Defaults to "SAP".
            env (str, optional): SAP environment. Defaults to "BW".
            output_file_extension (str, optional): Output file extension - to allow selection between .csv and .parquet. Defaults to ".parquet".
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Azure Data Lake file name. Defaults to None.
            adls_dir_path(str, optional): Azure Data Lake destination file path. Defaults to None.
            if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            vault_name (str, optional): The name of the vault from which to obtain the secrets.. Defaults to None.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
        """
        self.sapbw_credentials = sapbw_credentials
        self.sapbw_credentials_key = sapbw_credentials_key
        self.env = env
        self.mdx_query = mdx_query
        self.mapping_dict = mapping_dict
        self.output_file_extension = output_file_extension

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.now = str(pendulum.now("utc"))
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )

        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        self.if_exists = if_exists
        self.overwrite_adls = overwrite_adls
        self.vault_name = vault_name
        self.sp_credentials_secret = sp_credentials_secret

        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        sapbw_to_df_task = SAPBWToDF(
            sapbw_credentials=self.sapbw_credentials,
            sapbw_credentials_key=self.sapbw_credentials_key,
            env=self.env,
        )

        df = sapbw_to_df_task.bind(
            mdx_query=self.mdx_query,
            mapping_dict=self.mapping_dict,
            flow=self,
        )

        df_viadot_downloaded = add_ingestion_metadata_task.bind(df=df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_viadot_downloaded, flow=self)

        df_to_be_loaded = df_map_mixed_dtypes_for_parquet.bind(
            df_viadot_downloaded, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload()
        adls_upload = file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task = AzureDataLakeUpload()
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        df_viadot_downloaded.set_upstream(df, flow=self)
        dtypes_dict.set_upstream(df_viadot_downloaded, flow=self)
        df_to_be_loaded.set_upstream(dtypes_dict, flow=self)
        adls_upload.set_upstream(df_to_file, flow=self)

        df_to_file.set_upstream(dtypes_updated, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)

        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



from typing import List

import pandas as pd
import pandas_gbq
from google.oauth2 import service_account

from ..config import local_config
from ..exceptions import CredentialError, DBDataAccessError
from .base import Source
from ..utils import add_viadot_metadata_columns


class BigQuery(Source):
    """
    Class to connect with Bigquery project and SQL tables.

    Note that credentials used for authentication can be generated only for User Principal
    who has access to specific BigQuery project.
    """

    def __init__(
        self, credentials_key: str = None, credentials: dict = None, *args, **kwargs
    ):
        """
        Create an instance of BigQuery class.

        Args:
            credentials_key (str, optional): Credential key to dictionary where details are stored. Credentials can be generated as key
            for User Principal inside a BigQuery project. Defaults to None.
            credentials (str, optional): Credentials stored in dictionary. Defaults to None.

        Raises:
            CredentialError: In case credentials cannot be found.
        """
        DEFAULT_CREDENTIALS = local_config.get(credentials_key)
        credentials = credentials or DEFAULT_CREDENTIALS

        if credentials is None:
            raise CredentialError("Credentials not found.")

        credentials_service_account = (
            service_account.Credentials.from_service_account_info(credentials)
        )
        pandas_gbq.context.credentials = credentials_service_account

        super().__init__(*args, credentials=credentials, **kwargs)
        pandas_gbq.context.project = self.credentials["project_id"]

    def list_datasets(self) -> List[str]:
        """
        Get datasets from BigQuery project.

        Returns:
            List[str]: List of datasets from BigQuery project.
        """
        query = f"""SELECT schema_name 
                FROM {self.get_project_id()}.INFORMATION_SCHEMA.SCHEMATA
                """
        df = self.query_to_df(query)
        return df["schema_name"].values

    def list_tables(self, dataset_name: str) -> List[str]:
        """
        Get tables from BigQuery dataset. Dataset is required.

        Args:
            dataset_name (str): Dataset from Bigquery project.

        Returns:
            List[str]: List of tables from BigQuery dataset.
        """
        query = f"""SELECT table_name 
                FROM {self.get_project_id()}.{dataset_name}.INFORMATION_SCHEMA.TABLES
                """
        df = self.query_to_df(query)
        return df["table_name"].values

    def list_columns(self, dataset_name: str, table_name: str) -> List[str]:
        """
        Get columns from BigQuery table. Dataset name and Table name are required.

        Args:
            dataset_name (str): Dataset from Bigquery project.
            table_name (str): Table name from given dataset.
        Returns:
            List[str]: List of table names from the BigQuery dataset.
        """
        query = f"""SELECT column_name
                FROM {self.get_project_id()}.{dataset_name}.INFORMATION_SCHEMA.COLUMNS
                WHERE table_name="{table_name}"
                """
        df = self.query_to_df(query)
        return df["column_name"].values

    def get_project_id(self) -> str:
        """
        Get project id from json file generated for specific project.

        Returns:
            str: Project name.
        """
        return self.credentials["project_id"]

    @add_viadot_metadata_columns()
    def query_to_df(self, query: str) -> pd.DataFrame:
        """
        Query throught Bigquery table.
        `add_viadot_metadata_columns` adds one column to the DF - `_viadot_source`.

        Args:
            query (str): SQL-Like Query to return data values.

        Raises:
            DBDataAccessError: When dataset name or table name are incorrect.

        Returns:
            pd.DataFrame: Query result.
        """
        try:
            return pandas_gbq.read_gbq(query)
        except:
            raise DBDataAccessError



from viadot.sources import CloudForCustomers


def test_credentials():
    endpoint = "ServiceRequestCollection"
    c4c = CloudForCustomers(endpoint=endpoint, params={"$top": "2"})
    df = c4c.to_df(
        fields=["ProductRecipientPartyName", "CreationDateTime", "CreatedBy"]
    )
    assert df.empty == False



from unittest import mock

from viadot.flows.mysql_to_adls import MySqlToADLS

query = """SELECT * FROM `example-views`.`sales`"""


def test_instance_mysqltoadls():
    flow = MySqlToADLS("test_flow", country_short="DE")
    assert flow


def test_adls_gen1_to_azure_sql_new_mock(TEST_PARQUET_FILE_PATH):
    with mock.patch.object(MySqlToADLS, "run", return_value=True) as mock_method:
        flow = MySqlToADLS(
            "test_flow_de",
            country_short="DE",
            query=query,
            file_path=TEST_PARQUET_FILE_PATH,
            to_path=f"raw/examples/{TEST_PARQUET_FILE_PATH}",
            sp_credentials_secret="App-Azure-CR-DatalakeGen2-AIA-DEV",
            overwrite_adls=True,
        )
        flow.run()
        mock_method.assert_called_with()



import pandas as pd
from prefect.backend import get_key_value, set_key_value
from prefect.engine.state import Failed, Success
from prefect.tasks.secrets import PrefectSecret

from viadot.task_utils import custom_mail_state_handler, set_new_kv


def test_custom_state_handler():
    vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()

    final_state = custom_mail_state_handler(
        tracked_obj="Flow",
        old_state=Success,
        new_state=Failed,
        only_states=[Failed],
        local_api_key=None,
        credentials_secret="SENDGRIND",
        vault_name=vault_name,
    )

    assert final_state == Failed


def test_set_new_kv():
    df = pd.DataFrame(data={"col1": [1, 72, 24, 2], "col2": [0, 0, 3, 4]})
    set_new_kv.run(kv_name="test_for_setting_kv", df=df, filter_column="col1")
    result = get_key_value("test_for_setting_kv")
    assert result == "72"
    set_key_value(key="test_for_setting_kv", value=None)



from typing import Any, Dict, List, Literal, Union

import pandas as pd
from prefect import Flow, Task, apply_map

from viadot.task_utils import (
    add_ingestion_metadata_task,
    credentials_loader,
    df_to_csv,
    df_to_parquet,
    union_dfs_task,
)
from viadot.tasks import AzureDataLakeUpload, OutlookToDF


class OutlookToADLS(Flow):
    def __init__(
        self,
        mailbox_list: List[str],
        name: str = None,
        start_date: str = None,
        end_date: str = None,
        outbox_list: List[str] = ["Sent Items"],
        local_file_path: str = None,
        output_file_extension: str = ".parquet",
        adls_file_path: str = None,
        overwrite_adls: bool = True,
        adls_sp_credentials_secret: str = None,
        limit: int = 10000,
        timeout: int = 3600,
        if_exists: Literal["append", "replace", "skip"] = "append",
        outlook_credentials_secret: str = "OUTLOOK",
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from Outlook source to Azure Data Lake in parquet format by default.

        Args:
            mailbox_list (List[str]): Mailbox name.
            name (str, optional): The name of the flow. Defaults to None.
            start_date (str, optional): A filtering start date parameter e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A filtering end date parameter e.g. "2022-01-02". Defaults to None.
            outbox_list (List[str], optional): List of outbox folders to differenciate between
                Inboxes and Outboxes. Defaults to ["Sent Items"].
            local_file_path (str, optional): Local destination path. Defaults to None.
            output_file_extension (str, optional): Output file extension. Defaults to ".parquet".
            adls_file_path (str, optional): Azure Data Lake destination file path. Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake. Defaults to None.
            outlook_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with outlook credentials.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
            if_exists (Literal['append', 'replace', 'skip'], optional): What to do if the local file already exists. Defaults to "append".
        """

        self.mailbox_list = mailbox_list
        self.start_date = start_date
        self.end_date = end_date
        self.outbox_list = outbox_list
        self.limit = limit
        self.timeout = timeout
        self.local_file_path = local_file_path
        self.if_exsists = if_exists

        # AzureDataLakeUpload
        self.adls_file_path = adls_file_path
        self.output_file_extension = output_file_extension
        self.overwrite_adls = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.outlook_credentials_secret = outlook_credentials_secret

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_outlook_df(
        self, mailbox_list: Union[str, List[str]], flow: Flow = None
    ) -> Task:
        credentials = credentials_loader.run(
            credentials_secret=self.outlook_credentials_secret
        )
        outlook_to_df = OutlookToDF(timeout=self.timeout, credentials=credentials)
        df = outlook_to_df.bind(
            mailbox_name=mailbox_list,
            start_date=self.start_date,
            end_date=self.end_date,
            outbox_list=self.outbox_list,
            limit=self.limit,
            flow=flow,
        )

        return df

    def gen_flow(self) -> Flow:
        dfs = apply_map(self.gen_outlook_df, self.mailbox_list, flow=self)

        df = union_dfs_task.bind(dfs, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exsists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exsists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_with_metadata.set_upstream(df, flow=self)
        df_to_file.set_upstream(df_with_metadata, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)



import random

import pandas as pd
import pytest

from viadot.sources import CustomerGauge

ENDPOINT = random.choice(["responses", "non-responses"])
CG = CustomerGauge(endpoint=ENDPOINT)


def test_get_json_content():
    json_response = CG.get_json_response()
    assert isinstance(json_response, dict)
    assert isinstance(json_response["data"], list)
    assert isinstance(json_response["data"][2], dict)
    assert isinstance(json_response["cursor"], dict)


def test_properties_cleaning():
    json_response = CG.get_json_response()
    data = json_response["data"][2].copy()
    cleaned_data = CG.properties_cleaning(data.copy())
    assert isinstance(data["properties"], list)
    assert isinstance(cleaned_data["properties"], dict)


def test_flatten_json():
    nested_json = {
        "user": {
            "name": "Jane",
            "address": {
                "street": "456 Elm St",
                "city": "San Francisco",
                "state": "CA",
                "zip": "94109",
                "country": {"name": "United States", "code": "US"},
            },
            "phone_numbers": {"type": "home", "number": "555-4321"},
        }
    }

    expected_output = {
        "user_name": "Jane",
        "user_address_street": "456 Elm St",
        "user_address_city": "San Francisco",
        "user_address_state": "CA",
        "user_address_zip": "94109",
        "user_address_country_name": "United States",
        "user_address_country_code": "US",
        "user_phone_numbers_type": "home",
        "user_phone_numbers_number": "555-4321",
    }

    output = CG.flatten_json(nested_json)
    assert output == expected_output


def test_pagesize_and_to_df():
    json_response = CG.get_json_response(pagesize=1)
    df = CG.to_df(json_response)
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1


def test_pass_specific_cursor():
    # for default pagesize=1000 returned cursor value should be bigger than passed
    cur = random.randint(1, 9999)
    json_response = CG.get_json_response(cursor=cur)
    cur_retrieved = CG.get_cursor(json_response)
    assert cur_retrieved > cur


def test_uncomplete_date_arguments():
    with pytest.raises(ValueError, match="Missing date arguments"):
        json_response = CG.get_json_response(
            date_field="date_sent", start_date="2012-01-03"
        )


def test_endpoint_url_argument():
    ENDPOINT = random.choice(["responses", "non-responses"])
    ENDPOINT_URL = f"https://api.eu.customergauge.com/v7/rest/sync/{ENDPOINT}"
    CG = CustomerGauge(url=ENDPOINT_URL)
    json_response = CG.get_json_response()
    assert isinstance(json_response, dict)



import pytest

from viadot.sources import Supermetrics

RESPONSE_PIVOTED = {
    "meta": {
        "query": {
            "fields": [
                {
                    "id": "Date",
                    "field_id": "Date",
                    "field_name": "Date",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "profile",
                    "field_id": "profile",
                    "field_name": "View",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "segment",
                    "field_id": "segment",
                    "field_name": "Segment",
                    "field_type": "dim",
                    "field_split": "column",
                },
                {
                    "id": "Sessions",
                    "field_id": "Sessions",
                    "field_name": "Sessions",
                    "field_type": "met",
                    "field_split": "row",
                },
            ]
        },
        "result": {"total_columns": 6, "total_rows": 700},
    },
    "data": [
        [
            "Date",
            "View",
            "M-site_TOTAL: Bounces Landing",
            "M-site_TOTAL: Click to EDSP",
            "M-site_TOTAL: MQL Conversion Page Sessions",
            "M-site_TOTAL: Click to RWS",
        ],
        ["2020-01-01", "REDACTED", 123, 456, 78, 9],
    ],
}

RESPONSE_PIVOTED_NO_DATA = {
    "meta": {
        "query": {
            "fields": [
                {
                    "id": "Date",
                    "field_id": "Date",
                    "field_name": "Date",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "profileID",
                    "field_id": "profileID",
                    "field_name": "View ID",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "Hostname",
                    "field_id": "Hostname",
                    "field_name": "Hostname",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "profile",
                    "field_id": "profile",
                    "field_name": "View",
                    "field_type": "dim",
                    "field_split": "row",
                },
                {
                    "id": "segment",
                    "field_id": "segment",
                    "field_name": "Segment",
                    "field_type": "dim",
                    "field_split": "column",
                },
                {
                    "id": "Sessions",
                    "field_id": "Sessions",
                    "field_name": "Sessions",
                    "field_type": "met",
                    "field_split": "row",
                },
            ]
        },
        "result": {"total_columns": 0, "total_rows": 0},
    },
    "data": [],
}


def test___get_col_names_google_analytics_pivoted():
    columns = Supermetrics._get_col_names_google_analytics(response=RESPONSE_PIVOTED)
    assert columns == [
        "Date",
        "View",
        "M-site_TOTAL: Bounces Landing",
        "M-site_TOTAL: Click to EDSP",
        "M-site_TOTAL: MQL Conversion Page Sessions",
        "M-site_TOTAL: Click to RWS",
    ]


def test___get_col_names_google_analytics_pivoted_no_data():
    with pytest.raises(ValueError):
        Supermetrics._get_col_names_google_analytics(response=RESPONSE_PIVOTED_NO_DATA)



import json
from typing import Any, Dict

import prefect
from prefect import Task
from prefect.tasks.secrets import PrefectSecret

from viadot.config import local_config
from viadot.sources import AzureSQL

from .azure_key_vault import AzureKeyVaultSecret


class ASELiteToDF(Task):
    def __init__(
        self,
        credentials: Dict[str, Any] = None,
        query: str = None,
        timeout: int = 3600,
        *args,
        **kwargs
    ):
        """
        Task for obtaining data from ASElite source.
        Args:
            credentials (Dict[str, Any], optional): ASElite SQL Database credentials. Defaults to None.
            query(str, optional): Query to perform on a database. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        Returns: Pandas DataFrame
        """
        self.credentials = credentials
        self.query = query

        super().__init__(
            name="ASElite_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download from aselite database to df"""
        return super().__call__(*args, **kwargs)

    def run(
        self,
        query: str,
        credentials: Dict[str, Any] = None,
        credentials_secret: str = None,
        vault_name: str = None,
    ):
        logger = prefect.context.get("logger")
        if not credentials_secret:
            try:
                credentials_secret = PrefectSecret("aselite").run()
            except ValueError:
                pass

        if credentials_secret:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)
            logger.info("Loaded credentials from Key Vault")
        else:
            credentials = local_config.get("ASELite_SQL")
            logger.info("Loaded credentials from local source")

        aselite = AzureSQL(credentials=credentials)
        logger.info("Connected to ASELITE SOURCE")
        df = aselite.to_df(query=query)
        logger.info("Succefully collected data from query")
        return df



from viadot.tasks import LumaIngest


def test_luma_ingest():
    luma_task = LumaIngest(
        name="Luma test",
        metadata_dir_path="tests/resources/metadata/new",
        url="www.luma_url.com",
        dbt_project_path="project_path",
        credentials_secret="luma-dev",
    )
    assert luma_task.name == "Luma test"
    assert isinstance(luma_task, LumaIngest)



from typing import Any, Dict, List

from prefect import Flow
from prefect.utilities import logging

from ..tasks import AzureDataLakeDownload, BlobFromCSV, CreateTableFromBlob

logger = logging.get_logger(__name__)


class ADLSGen1ToAzureSQL(Flow):
    """Bulk insert a file from an Azure Data Lake gen1 to Azure SQL Database.

    Args:
        name (str): The name of the flow.
        path (str): The path to the Data Lake file/folder.
        blob_path (str): The path of the generated blob.
        dtypes (dict): Which dtypes to use when creating the table in Azure SQL Database.
        local_file_path (str): Where the gen1 file should be downloaded.
        sp_credentials_secret (str): The Key Vault secret holding Service Pricipal credentials
        vault_name (str): The name of the vault from which to retrieve `sp_credentials_secret`
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        name: str,
        path: str,
        blob_path: str,
        local_file_path: str = None,
        dtypes: Dict[str, Any] = None,
        overwrite_blob: bool = True,
        sep: str = "\t",
        table: str = None,
        schema: str = None,
        if_exists: str = "replace",
        sp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.path = path
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.blob_path = blob_path
        self.overwrite_blob = overwrite_blob
        self.sep = sep
        self.table = table
        self.schema = schema
        self.dtypes = dtypes
        self.if_exists = if_exists
        self.sp_credentials_secret = sp_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        gen1_download_task = AzureDataLakeDownload(gen=1, timeout=self.timeout)
        gen1_download_task.bind(
            from_path=self.path,
            to_path=self.local_file_path,
            gen=1,
            sp_credentials_secret=self.sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        csv_to_blob_storage_task = BlobFromCSV(timeout=self.timeout)
        csv_to_blob_storage_task.bind(
            from_path=self.local_file_path,
            to_path=self.blob_path,
            overwrite=self.overwrite_blob,
            flow=self,
        )
        blob_to_azure_sql_task = CreateTableFromBlob(timeout=self.timeout)
        blob_to_azure_sql_task.bind(
            blob_path=self.blob_path,
            schema=self.schema,
            table=self.table,
            dtypes=self.dtypes,
            sep=self.sep,
            if_exists=self.if_exists,
            flow=self,
        )

        csv_to_blob_storage_task.set_upstream(gen1_download_task, flow=self)
        blob_to_azure_sql_task.set_upstream(csv_to_blob_storage_task, flow=self)



import os

import pytest

from viadot.sources.duckdb import DuckDB
from viadot.tasks import DuckDBCreateTableFromParquet

TABLE = "test_table"
SCHEMA = "test_schema"
DATABASE_PATH = "test_db_123.duckdb"


@pytest.fixture(scope="module")
def duckdb():
    duckdb = DuckDB(credentials=dict(database=DATABASE_PATH))
    yield duckdb


def test_create_table_empty_file(duckdb):
    path = "empty.parquet"
    with open(path, "w"):
        pass
    duckdb_creds = {f"database": DATABASE_PATH}
    task = DuckDBCreateTableFromParquet(credentials=duckdb_creds)
    task.run(schema=SCHEMA, table=TABLE, path=path, if_empty="skip")

    assert duckdb._check_if_table_exists(TABLE, schema=SCHEMA) == False
    os.remove(path)


def test_create_table(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb_creds = {f"database": DATABASE_PATH}
    task = DuckDBCreateTableFromParquet(credentials=duckdb_creds)
    task.run(schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH, if_empty="skip")

    assert duckdb._check_if_table_exists(TABLE, schema=SCHEMA)
    os.remove(DATABASE_PATH)



import pandas as pd
import pytest

from viadot.tasks.hubspot import HubspotToDF


@pytest.fixture(scope="session")
def var_dictionary():
    variables = {
        "endpoint": "contacts",
        "filters": [
            {
                "filters": [
                    {
                        "propertyName": "createdate",
                        "operator": "BETWEEN",
                        "highValue": "1642636800000",
                        "value": "1642535000000",
                    },
                    {
                        "propertyName": "email",
                        "operator": "CONTAINS_TOKEN",
                        "value": "*@seznam.cz",
                    },
                ]
            }
        ],
        "properties": [
            "recent_conversion_date",
            "recent_conversion_event_name",
            "hs_analytics_num_visits",
        ],
    }
    yield variables


def test_hubspot_to_df(var_dictionary):
    endpoint = var_dictionary["endpoint"]
    properties = var_dictionary["properties"]
    filters = var_dictionary["filters"]
    nrows = 150

    hubspot_to_df_task = HubspotToDF(hubspot_credentials_key="HUBSPOT")

    df = hubspot_to_df_task.run(
        endpoint=endpoint, properties=properties, filters=filters, nrows=nrows
    )

    assert isinstance(df, pd.DataFrame)



from typing import Any, Dict, List

import pandas as pd
from prefect import Flow
from prefect.utilities import logging
from prefect.utilities.tasks import task

from viadot.flows.adls_to_azure_sql import df_to_csv_task
from viadot.task_utils import METADATA_COLUMNS, add_ingestion_metadata_task

from ..tasks import AzureDataLakeToDF, AzureDataLakeUpload, AzureSQLCreateTable, BCPTask

logger = logging.get_logger(__name__)


@task
def df_to_csv_task(df: pd.DataFrame, path: str, sep: str = "\t") -> None:
    df.to_csv(path, index=False, sep=sep)


@task
def df_replace_special_chars(df: pd.DataFrame) -> None:
    df = df.replace(r"\n|\t", "", regex=True)
    return df


class ADLSGen1ToAzureSQLNew(Flow):
    """Move file(s) from Azure Data Lake gen1 to gen2.

    Args:
        name (str): The name of the flow.
        gen1_path (str): The path to the gen1 Data Lake file/folder.
        gen2_path (str): The path of the final gen2 file/folder.
        local_file_path (str): Where the gen1 file should be downloaded.
        overwrite (str): Whether to overwrite the destination file(s).
        read_sep (str): The delimiter for the gen1 file.
        write_sep (str): The delimiter for the output file.
        read_quoting (str): The quoting option for the input file.
        read_lineterminator (str): The line terminator for the input file.
        read_error_bad_lines (bool): Whether to raise an exception on bad lines.
        gen1_sp_credentials_secret (str): The Key Vault secret holding Service Pricipal credentials for gen1 lake
        gen2_sp_credentials_secret (str): The Key Vault secret holding Service Pricipal credentials for gen2 lake
        sqldb_credentials_secret (str): The Key Vault secret holding Azure SQL Database credentials
        vault_name (str): The name of the vault from which to retrieve `sp_credentials_secret`
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        name: str,
        gen1_path: str,
        gen2_path: str,
        local_file_path: str = None,
        overwrite: bool = True,
        read_sep: str = "\t",
        write_sep: str = "\t",
        read_quoting: str = None,
        read_lineterminator: str = None,
        read_error_bad_lines: bool = True,
        schema: str = None,
        table: str = None,
        dtypes: dict = None,
        if_exists: str = "fail",
        gen1_sp_credentials_secret: str = None,
        gen2_sp_credentials_secret: str = None,
        sqldb_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.gen1_path = gen1_path
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.gen2_path = gen2_path
        self.overwrite = overwrite
        self.read_sep = read_sep
        self.write_sep = write_sep
        self.read_quoting = read_quoting
        self.read_lineterminator = read_lineterminator
        self.read_error_bad_lines = read_error_bad_lines
        self.schema = schema
        self.table = table
        self.dtypes = dtypes
        self.if_exists = if_exists
        self.gen1_sp_credentials_secret = gen1_sp_credentials_secret
        self.gen2_sp_credentials_secret = gen2_sp_credentials_secret
        self.sqldb_credentials_secret = sqldb_credentials_secret
        self.vault_name = vault_name
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.dtypes.update(METADATA_COLUMNS)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        gen1_to_df_task = AzureDataLakeToDF(gen=1, timeout=self.timeout)
        df = gen1_to_df_task.bind(
            path=self.gen1_path,
            gen=1,
            lineterminator=self.read_lineterminator,
            sep=self.read_sep,
            quoting=self.read_quoting,
            error_bad_lines=self.read_error_bad_lines,
            sp_credentials_secret=self.gen1_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        df2 = df_replace_special_chars.bind(df=df, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df=df2, flow=self)
        df_to_csv_task.bind(
            df=df_with_metadata,
            path=self.local_file_path,
            sep=self.write_sep,
            flow=self,
        )
        gen2_upload_task = AzureDataLakeUpload(gen=2, timeout=self.timeout)
        gen2_upload_task.bind(
            from_path=self.local_file_path,
            to_path=self.gen2_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.gen2_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        create_table_task = AzureSQLCreateTable(timeout=self.timeout)
        create_table_task.bind(
            schema=self.schema,
            table=self.table,
            dtypes=self.dtypes,
            if_exists=self.if_exists,
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )
        bulk_insert_task = BCPTask(timeout=self.timeout)
        bulk_insert_task.bind(
            path=self.local_file_path,
            schema=self.schema,
            table=self.table,
            credentials_secret=self.sqldb_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        df_with_metadata.set_upstream(df_replace_special_chars, flow=self)
        df_to_csv_task.set_upstream(df_with_metadata, flow=self)
        gen2_upload_task.set_upstream(df_to_csv_task, flow=self)
        create_table_task.set_upstream(df_to_csv_task, flow=self)
        bulk_insert_task.set_upstream(create_table_task, flow=self)



import os

import pytest

from viadot.sources import CloudForCustomers

TEST_FILE_1 = "tests_out.csv"


@pytest.fixture(scope="session")
def cloud_for_customers():
    url = "http://services.odata.org/V2/Northwind/Northwind.svc/"
    endpoint = "Employees"
    cloud_for_customers = CloudForCustomers(
        url=url, endpoint=endpoint, params={"$top": "2"}
    )
    yield cloud_for_customers
    os.remove(TEST_FILE_1)


def test_to_df(cloud_for_customers):
    df = cloud_for_customers.to_df(fields=["EmployeeID", "FirstName", "LastName"])
    assert not df.empty
    assert df.shape[1] == 3
    assert df.shape[0] == 2


def test_to_records(cloud_for_customers):
    data = cloud_for_customers.to_records()
    assert "EmployeeID" in data[0].keys()
    assert data != []
    assert len(data) == 2


def test_csv(cloud_for_customers):
    csv = cloud_for_customers.to_csv(path=TEST_FILE_1)
    assert os.path.isfile(TEST_FILE_1) == True



class ValidationError(Exception):
    pass


class APIError(Exception):
    pass


class CredentialError(Exception):
    pass


class DBDataAccessError(Exception):
    pass


class DataRangeError(Exception):
    pass


class DataBufferExceeded(Exception):
    pass



from typing import Any, Dict, List, Literal

from prefect import Flow, config

from viadot.tasks import SQLServerQuery


class SQLServerTransform(Flow):
    def __init__(
        self,
        name: str,
        query: str,
        config_key: str,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for running SQL queries on top of SQL Server.

        Args:
            name (str,required): The name of the flow.
            query (str, required): The query to execute on the database.
            config_key (str, required): Config key containing credentials for the SQL Server connection.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.query = query
        self.config_key = config_key
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    def gen_flow(self) -> Flow:
        query_task = SQLServerQuery(timeout=self.timeout)
        query_task.bind(
            query=self.query,
            config_key=self.config_key,
            flow=self,
        )



from .aselite import ASELiteToDF
from .azure_blob_storage import BlobFromCSV
from .azure_data_lake import (
    AzureDataLakeCopy,
    AzureDataLakeDownload,
    AzureDataLakeList,
    AzureDataLakeRemove,
    AzureDataLakeToDF,
    AzureDataLakeUpload,
)
from .azure_key_vault import (
    AzureKeyVaultSecret,
    CreateAzureKeyVaultSecret,
    DeleteAzureKeyVaultSecret,
)
from .azure_sql import (
    AzureSQLBulkInsert,
    AzureSQLCreateTable,
    AzureSQLDBQuery,
    AzureSQLToDF,
    AzureSQLUpsert,
    CheckColumnOrder,
    CreateTableFromBlob,
)
from .bcp import BCPTask
from .bigquery import BigQueryToDF
from .cloud_for_customers import C4CReportToDF, C4CToDF
from .genesys import GenesysToCSV
from .github import DownloadGitHubFile
from .great_expectations import RunGreatExpectationsValidation
from .outlook import OutlookToDF
from .prefect_date_range import GetFlowNewDateRange
from .salesforce import SalesforceBulkUpsert, SalesforceToDF, SalesforceUpsert
from .sharepoint import SharepointToDF
from .sqlite import SQLiteInsert, SQLiteQuery, SQLiteSQLtoDF
from .supermetrics import SupermetricsToCSV, SupermetricsToDF

try:
    from .sap_rfc import SAPRFCToDF
except ImportError:
    pass

try:
    from .sap_bw import SAPBWToDF
except ImportError:
    pass

from .business_core import BusinessCoreToParquet
from .customer_gauge import CustomerGaugeToDF
from .duckdb import DuckDBCreateTableFromParquet, DuckDBQuery, DuckDBToDF
from .epicor import EpicorOrdersToDF
from .eurostat import EurostatToDF
from .hubspot import HubspotToDF
from .mediatool import MediatoolToDF
from .mindful import MindfulToCSV
from .sftp import SftpList, SftpToDF
from .sql_server import SQLServerCreateTable, SQLServerQuery, SQLServerToDF
from .vid_club import VidClubToDF
from .git import CloneRepo
from .luma import LumaIngest


from typing import Any, Dict, List, Literal

from prefect import Flow

from viadot.task_utils import add_ingestion_metadata_task, cast_df_to_str, df_to_parquet
from viadot.tasks import DuckDBCreateTableFromParquet, SQLServerToDF


class SQLServerToDuckDB(Flow):
    def __init__(
        self,
        name,
        sql_query: str,
        local_file_path: str,
        sqlserver_config_key: str = None,
        duckdb_table: str = None,
        duckdb_schema: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
        if_empty: Literal["warn", "skip", "fail"] = "skip",
        duckdb_credentials: dict = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for upolading data from SQL Server to DuckDB.

        Args:
            name (str): The name of the flow.
            sql_query (str, required): The query to execute on the SQL Server database. If don't start with "SELECT"
                returns empty DataFrame.
            local_file_path (str): Path to output parquet file.
            sqlserver_config_key (str, optional): The key inside local config containing the credentials. Defaults to None.
            duckdb_table (str, optional): Destination table in DuckDB. Defaults to None.
            duckdb_schema (str, optional): Destination schema in DuckDB. Defaults to None.
            if_exists (Literal, optional):  What to do if the table already exists. Defaults to "fail".
            if_empty (Literal, optional): What to do if Parquet file is empty. Defaults to "skip".
            duckdb_credentials (dict, optional): Credentials for the DuckDB connection. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SQLServerToDF
        self.sql_query = sql_query
        self.sqlserver_config_key = sqlserver_config_key
        self.timeout = timeout

        # DuckDBCreateTableFromParquet
        self.local_file_path = local_file_path
        self.duckdb_table = duckdb_table
        self.duckdb_schema = duckdb_schema
        self.if_exists = if_exists
        self.if_empty = if_empty
        self.duckdb_credentials = duckdb_credentials

        super().__init__(*args, name=name, **kwargs)

        self.create_duckdb_table_task = DuckDBCreateTableFromParquet(
            credentials=duckdb_credentials, timeout=timeout
        )

        self.gen_flow()

    def gen_flow(self) -> Flow:
        df_task = SQLServerToDF(timeout=self.timeout)
        df = df_task.bind(
            config_key=self.sqlserver_config_key, query=self.sql_query, flow=self
        )
        df_mapped = cast_df_to_str.bind(df, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df_mapped, flow=self)
        parquet = df_to_parquet.bind(
            df=df_with_metadata,
            path=self.local_file_path,
            if_exists=self.if_exists,
            flow=self,
        )
        create_duckdb_table = self.create_duckdb_table_task.bind(
            path=self.local_file_path,
            schema=self.duckdb_schema,
            table=self.duckdb_table,
            if_exists=self.if_exists,
            if_empty=self.if_empty,
            flow=self,
        )
        create_duckdb_table.set_upstream(parquet, flow=self)



import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging

from viadot.exceptions import ValidationError
from viadot.sources import SAPBW
from viadot.task_utils import *

logger = logging.get_logger()


class SAPBWToDF(Task):
    def __init__(
        self,
        sapbw_credentials: dict = None,
        sapbw_credentials_key: str = "SAP",
        env: str = "BW",
        *args,
        **kwargs,
    ):
        """
        A task for quering the SAP BW (SAP Business Warehouse) source using pyrfc library.

        Args:
            sapbw_credentials (dict, optional): Credentials to SAP BW server. Defaults to None.
            sapbw_credentials_key (str, optional): Azure KV secret. Defaults to "SAP".
            env (str, optional): SAP environment. Defaults to "BW".
        """
        if sapbw_credentials is None:
            self.sapbw_credentials = credentials_loader.run(
                credentials_secret=sapbw_credentials_key
            ).get(env)

        else:
            self.sapbw_credentials = sapbw_credentials

        super().__init__(
            name="sapbw_to_df",
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download SAP BW data to a DF"""
        super().__call__(self)

    def apply_user_mapping(
        self, df: pd.DataFrame, mapping_dict: dict = {}
    ) -> pd.DataFrame:
        """
        Function to apply the column mapping defined by user for the output dataframe.
        DataFrame will be cut to selected columns - if any other columns need to be included in the output file,
        please add them to the mapping dictionary with original names.

        Args:
            df (pd.DataFrame): Input dataframe for the column mapping task.
            mapping_dict (dict, optional): Dictionary with original and new column names. Defaults to {}.

        Returns:
            pd.DataFrame: Output DataFrame with mapped columns.
        """
        self.logger.info("Applying user defined mapping for columns...")
        df = df[mapping_dict.keys()]
        df.columns = mapping_dict.values()

        self.logger.info(f"Successfully applied user mapping.")

        return df

    def to_df(self, query_output: dict) -> pd.DataFrame:
        """
        Function to convert the SAP output in JSON format into a dataframe.

        Args:
            query_output (dict): SAP output dictionary in JSON format that contains data rows and column headers.

        Raises:
            ValidationError: Prints the original SAP error message in case of issues with MDX execution.

        Returns:
            pd.DataFrame: Output dataframe.
        """
        raw_data = {}

        if query_output["RETURN"]["MESSAGE"] == "":
            results = query_output["DATA"]
            for cell in results:
                if cell["ROW"] not in raw_data:
                    raw_data[cell["ROW"]] = {}
                if "].[" not in cell["DATA"]:
                    raw_data[cell["ROW"]][cell["COLUMN"]] = cell["DATA"]
            rows = [raw_data[row] for row in raw_data]
            cols = [x["DATA"] for x in query_output["HEADER"]]
            df = pd.DataFrame(data=rows)
            df.columns = cols
        else:
            df = pd.DataFrame()
            raise ValidationError(query_output["RETURN"]["MESSAGE"])

        return df

    def run(self, mdx_query: str, mapping_dict: dict = {}) -> pd.DataFrame:
        """
        Task run method.

        Args:
            mdx_query (str): MDX query to be passed to SAP BW.
            mapping_dict (dict, optional): Mapping dictionary from user in json format. Defaults to {}.

        Returns:
            pd.DataFrame: Output DataFrame with applied column mapping.
        """
        sap = SAPBW(credentials=self.sapbw_credentials)

        data = sap.get_output_data(mdx_query)
        df = self.to_df(data)

        if mapping_dict:
            df = self.apply_user_mapping(df, mapping_dict)

        return df



from typing import Any, Dict

import sharepy

from ..config import local_config
from ..exceptions import CredentialError
from .base import Source


class Sharepoint(Source):
    """
    A Sharepoint class to connect and download specific Excel file from Sharepoint.

    Args:
        credentials (dict): In credentials should be included:
            "site" - Path to sharepoint website (e.g : {tenant_name}.sharepoint.com)
            "username" - Sharepoint username (e.g username@{tenant_name}.com)
            "password"
        download_from_path (str, optional): Full url to file
                        (e.g : https://{tenant_name}.sharepoint.com/sites/{directory}/Shared%20Documents/Dashboard/file). Defaults to None.
    """

    def __init__(
        self,
        credentials: Dict[str, Any] = None,
        download_from_path: str = None,
        *args,
        **kwargs,
    ):
        DEFAULT_CREDENTIALS = local_config.get("SHAREPOINT")
        credentials = credentials or DEFAULT_CREDENTIALS
        if credentials is None:
            raise CredentialError("Credentials not found.")
        self.url = download_from_path
        self.required_credentials = ["site", "username", "password"]
        super().__init__(*args, credentials=credentials, **kwargs)

    def get_connection(self) -> sharepy.session.SharePointSession:
        if any([rq not in self.credentials for rq in self.required_credentials]):
            raise CredentialError("Missing credentials.")

        return sharepy.connect(
            site=self.credentials["site"],
            username=self.credentials["username"],
            password=self.credentials["password"],
        )

    def download_file(
        self,
        download_from_path: str = None,
        download_to_path: str = "Sharepoint_file.xlsm",
    ) -> None:
        """Function to download files from Sharepoint.
        Args:
            download_from_path (str): Path from which to download file. Defaults to None.
            download_to_path (str, optional): Path to destination file. Defaults to "Sharepoint_file.xlsm".
        """
        download_from_path = download_from_path or self.url
        if not download_from_path:
            raise ValueError("Missing required parameter 'download_from_path'.")

        conn = self.get_connection()
        conn.getfile(
            url=download_from_path,
            filename=download_to_path,
        )



from datetime import timedelta
from typing import List

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

try:
    from viadot.sources import SAPRFC, SAPRFCV2
except ImportError:
    raise


class SAPRFCToDF(Task):
    def __init__(
        self,
        query: str = None,
        sep: str = None,
        replacement: str = "-",
        func: str = None,
        rfc_total_col_width_character_limit: int = 400,
        credentials: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        A task for querying SAP with SQL using the RFC protocol.

        Note that only a very limited subset of SQL is supported:
        - aliases
        - where clauses combined using the AND operator
        - limit & offset

        Unsupported:
        - aggregations
        - joins
        - subqueries
        - etc.

        Args:
            query (str, optional): The query to be executed with pyRFC.
            sep (str, optional): The separator to use when reading query results. If not provided,
            multiple options are automatically tried. Defaults to None.
            replacement (str, optional): In case of sep is on a columns, set up a new character to replace
                inside the string to avoid flow breakdowns. Defaults to "-".
            func (str, optional): SAP RFC function to use. Defaults to None.
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
            in case of too many columns for RFC function. According to SAP documentation, the limit is
            512 characters. However, we observed SAP raising an exception even on a slightly lower number
            of characters, so we add a safety margin. Defaults to 400.
            credentials (dict, optional): The credentials to use to authenticate with SAP.
            By default, they're taken from the local viadot config.
        """
        self.query = query
        self.sep = sep
        self.replacement = replacement
        self.credentials = credentials
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit

        super().__init__(
            name="sap_rfc_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "query",
        "sep",
        "replacement",
        "func",
        "rfc_total_col_width_character_limit",
        "credentials",
    )
    def run(
        self,
        query: str = None,
        sep: str = None,
        replacement: str = "-",
        credentials: dict = None,
        func: str = None,
        rfc_total_col_width_character_limit: int = None,
        rfc_unique_id: List[str] = None,
        alternative_version: bool = False,
    ) -> pd.DataFrame:
        """Task run method.

        Args:
            query (str, optional): The query to be executed with pyRFC.
            sep (str, optional): The separator to use when reading query results. If not provided,
            multiple options are automatically tried. Defaults to None.
            replacement (str, optional): In case of sep is on a columns, set up a new character to replace
                inside the string to avoid flow breakdowns. Defaults to "-".
            func (str, optional): SAP RFC function to use. Defaults to None.
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
                in case of too many columns for RFC function. According to SAP documentation, the limit is
                512 characters. However, we observed SAP raising an exception even on a slightly lower number
                of characters, so we add a safety margin. Defaults to None.
            rfc_unique_id  (List[str], optional): Reference columns to merge chunks Data Frames. These columns must to be unique. If no columns are provided
                in this parameter, all data frame columns will by concatenated. Defaults to None.
                Example:
                --------
                SAPRFCToADLS(
                    ...
                    rfc_unique_id=["VBELN", "LPRIO"],
                    ...
                    )
            alternative_version (bool, optional): Enable the use version 2 in source. Defaults to False.

        Returns:
            pd.DataFrame: DataFrame with SAP data.
        """
        if query is None:
            raise ValueError("Please provide the query.")

        if alternative_version is True:
            if rfc_unique_id:
                self.logger.warning(
                    "If the column/set of columns are not unique the table will be malformed."
                )
            sap = SAPRFCV2(
                sep=sep,
                replacement=replacement,
                credentials=credentials,
                func=func,
                rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
                rfc_unique_id=rfc_unique_id,
            )
        else:
            sap = SAPRFC(
                sep=sep,
                credentials=credentials,
                func=func,
                rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
            )
        sap.query(query)
        self.logger.info(f"Downloading data from SAP to a DataFrame...")
        self.logger.debug(f"Running query: \n{query}.")

        df = sap.to_df()

        self.logger.info(f"Data has been downloaded successfully.")
        return df



import os
import uuid
from unittest import mock

import pytest

from viadot.sources import AzureDataLake
from viadot.tasks import (
    AzureDataLakeCopy,
    AzureDataLakeDownload,
    AzureDataLakeList,
    AzureDataLakeRemove,
    AzureDataLakeToDF,
    AzureDataLakeUpload,
)

uuid_4 = uuid.uuid4()
uuid_4_2 = uuid.uuid4()

FILE_NAME = f"test_file_{uuid_4}.csv"
FILE_NAME_2 = f"test_file_{uuid_4}.csv"
ADLS_PATH = f"raw/supermetrics/{FILE_NAME}"
ADLS_PATH_2 = f"raw/supermetrics/{FILE_NAME_2}"

FILE_NAME_PARQUET = f"test_file_{uuid_4}.parquet"
ADLS_PATH_PARQUET = f"raw/supermetrics/{FILE_NAME_PARQUET}"

ADLS_TEST_PATHS = [
    "raw/tests/alds_test_new_fnc/2020/02/01/final_df2.csv",
    "raw/tests/alds_test_new_fnc/2020/02/01/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/2020/02/02/final_df2.csv",
    "raw/tests/alds_test_new_fnc/2020/02/02/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/2021/12/01/final_df2.csv",
    "raw/tests/alds_test_new_fnc/2021/12/01/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/2022/06/21/final_df2.csv",
    "raw/tests/alds_test_new_fnc/2022/06/21/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/2022/08/12/final_df2.csv",
    "raw/tests/alds_test_new_fnc/2022/08/12/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/test_folder/final_df2.csv",
    "raw/tests/alds_test_new_fnc/test_folder/test_new_fnc.csv",
    "raw/tests/alds_test_new_fnc/test_folder_2/final_df2.csv",
    "raw/tests/alds_test_new_fnc/test_folder_2/test_new_fnc.csv",
]


@pytest.mark.dependency()
def test_azure_data_lake_upload(TEST_CSV_FILE_PATH):
    upload_task = AzureDataLakeUpload()
    upload_task.run(from_path=TEST_CSV_FILE_PATH, to_path=ADLS_PATH)
    file = AzureDataLake(ADLS_PATH)
    assert file.exists()


@pytest.mark.dependency(depends=["test_azure_data_lake_upload"])
def test_azure_data_lake_download():
    download_task = AzureDataLakeDownload()
    download_task.run(from_path=ADLS_PATH)
    assert os.path.exists(FILE_NAME)
    os.remove(FILE_NAME)


@pytest.mark.dependency(depends=["test_azure_data_lake_upload"])
def test_azure_data_lake_to_df():
    task = AzureDataLakeToDF()
    df = task.run(path=ADLS_PATH, sep="\t")
    assert not df.empty


@pytest.mark.dependency(
    depends=["test_azure_data_lake_upload", "test_azure_data_lake_to_df"]
)
def test_azure_data_lake_to_df_parquet(TEST_PARQUET_FILE_PATH):
    upload_task = AzureDataLakeUpload()
    upload_task.run(from_path=TEST_PARQUET_FILE_PATH, to_path=ADLS_PATH_PARQUET)

    lake_to_df_task = AzureDataLakeToDF()
    df = lake_to_df_task.run(path=ADLS_PATH_PARQUET)
    assert not df.empty


@pytest.mark.dependency(depends=["test_azure_data_lake_upload"])
def test_azure_data_lake_copy():
    copy_task = AzureDataLakeCopy()
    copy_task.run(from_path=ADLS_PATH, to_path=ADLS_PATH_2)
    file = AzureDataLake(ADLS_PATH_2)
    assert file.exists()


def test_azure_data_lake_list():
    list_task = AzureDataLakeList()
    files = list_task.run(path="raw/supermetrics")
    assert ADLS_PATH in files


def test_azure_data_lake_list_recursive():
    list_task = AzureDataLakeList()
    files = list_task.run(path="raw/tests/alds_test_new_fnc/", recursive=True)
    assert isinstance(files, list)


def test_azure_data_lake_list_paths():
    with mock.patch.object(
        AzureDataLakeList, "run", return_value=ADLS_TEST_PATHS
    ) as mock_method:
        list_task = AzureDataLakeList(path="raw/tests/alds_test_new_fnc/")
        files = list_task.run(recursive=True)
        assert files == ADLS_TEST_PATHS


@pytest.mark.dependency(depends=["test_azure_data_lake_upload"])
def test_azure_data_lake_remove():
    file = AzureDataLake(ADLS_PATH)
    assert file.exists()
    remove_task = AzureDataLakeRemove()
    remove_task.run(path=ADLS_PATH)
    assert not file.exists()



from ..config import local_config
from ..flows import SupermetricsToAzureSQLv3

SUPERMETRICS_CREDENTIALS = local_config.get("SUPERMETRICS")

# Note this flow uses local config, as well as Prefect and Azure Key Vault
# secrets for the Key Vault, ADLS, and Azure SQL Database.
# These defaults are set with the following local Prefect secrets:
# AZURE_CREDENTIALS, AZURE_DEFAULT_KEYVAULT,
# AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET,
# and AZURE_DEFAULT_SQLDB_SERVICE_PRINCIPAL_SECRET.
# Make sure to specify them in your .prefect/config.toml file.
# For details on each secret, see the relevant task's documentation.

google_ads_flow = SupermetricsToAzureSQLv3(
    "Google Ads extract",
    ds_id="AW",
    ds_accounts=SUPERMETRICS_CREDENTIALS["SOURCES"]["Google Ads"]["Accounts"],
    ds_user=SUPERMETRICS_CREDENTIALS["USER"],
    date_range_type="last_year_inc",
    fields=[
        "Date",
        "profile",
        "Campaignname",
        "Impressions",
        "Clicks",
        "Cost_eur",
        "SearchImpressionShare",
    ],
    max_rows=1000000,
    dtypes={
        "Date": "DATE",
        "Account": "VARCHAR(255)",
        "Campaign": "VARCHAR(255)",
        "Impressions": "FLOAT(24)",
        "Clicks": "FLOAT(24)",
        "Cost": "FLOAT(24)",
        "ImpressionShare": "VARCHAR(255)",
    },
    schema="sandbox",
    table="google_ads_test",
    adls_path="tests/supermetrics/google_ads.csv",
)

# google_ads_flow.visualize()  # print the generated execution graph
google_ads_flow.run()



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import CustomerGaugeToADLS

DATA = {
    "user_name": ["Jane", "Bob"],
    "user_address_street": ["456 Elm St", "55 Oat St"],
    "user_address_city": ["San Francisco", "New York City"],
    "user_address_state": ["CA", "NY"],
    "user_address_zip": ["94109", "85674"],
    "user_address_country_name": "United States",
    "user_address_country_code": "US",
}
COLUMNS = ["user_name", "user_address_street"]
ADLS_FILE_NAME = "test_customer_gauge.parquet"
ADLS_DIR_PATH = "raw/tests/"


@mock.patch(
    "viadot.tasks.CustomerGaugeToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_customer_gauge_to_adls_run_flow(mocked_class):
    flow = CustomerGaugeToADLS(
        "test_customer_gauge_to_adls_flow_run",
        endpoint="responses",
        total_load=False,
        anonymize=True,
        columns_to_anonymize=COLUMNS,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
        overwrite_adls=True,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_customer_gauge_to_adls_flow_run.parquet")
    os.remove("test_customer_gauge_to_adls_flow_run.json")



import logging

import pandas as pd
import pytest

from viadot.tasks import eurostat


def test_and_validate_dataset_code_without_params(caplog):
    """This function is designed to test the accuracy of the data retrieval feature in a program.
    Specifically, it tests to ensure that the program returns a non-empty DataFrame when a correct
    dataset code is provided without any parameters. The function is intended to be used in software
    development to verify that the program is correctly retrieving data from the appropriate dataset.
    """
    task = eurostat.EurostatToDF(dataset_code="ILC_DI04").run()
    assert isinstance(task, pd.DataFrame)
    assert not task.empty
    assert caplog.text == ""


def test_wrong_dataset_code_logger(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log errors
    when provided with only incorrect dataset code.
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    task = eurostat.EurostatToDF(dataset_code="ILC_DI04E")

    with pytest.raises(ValueError, match="DataFrame is empty!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        f"Failed to fetch data for ILC_DI04E, please check correctness of dataset code!"
        in caplog.text
    )


def test_wrong_parameters_codes_logger(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log errors
    when provided with a correct dataset_code and correct parameters are provided, but both parameters codes are incorrect.
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04",
        params={"hhtyp": "total1", "indic_il": "non_existing_code"},
    )

    with pytest.raises(ValueError, match="DataFrame is empty!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        f"Parameters codes: 'total1 | non_existing_code' are not available. Please check your spelling!"
        in caplog.text
    )
    assert (
        f"You can find everything via link: https://ec.europa.eu/eurostat/databrowser/view/ILC_DI04/default/table?lang=en"
        in caplog.text
    )


def test_parameter_codes_as_list_logger(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log errors
    when provided with a correct dataset code, correct parameters, but incorrect parameters codes structure
    (as a list with strings, instead of single string).
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """

    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04",
        params={"hhtyp": ["totale", "nottotale"], "indic_il": "med_e"},
    )
    with pytest.raises(ValueError, match="Wrong structure of params!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        "You can provide only one code per one parameter as 'str' in params!\n"
        in caplog.text
    )
    assert (
        "CORRECT: params = {'unit': 'EUR'} | INCORRECT: params = {'unit': ['EUR', 'USD', 'PLN']}"
        in caplog.text
    )


def test_wrong_parameters(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log errors
    when provided with a correct dataset_code, but incorrect parameters keys.
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """

    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04", params={"hhhtyp": "total", "indic_ilx": "med_e"}
    )
    with pytest.raises(ValueError, match="DataFrame is empty!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        f"Parameters: 'hhhtyp | indic_ilx' are not in dataset. Please check your spelling!\n"
        in caplog.text
    )
    assert (
        f"Possible parameters: freq | hhtyp | indic_il | unit | geo | time"
        in caplog.text
    )


def test_params_as_list():
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log error
    when provided with a correct dataset_code, but incorrect params structure (as list instead of dict).
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    with pytest.raises(TypeError, match="Params should be a dictionary."):
        eurostat.EurostatToDF(dataset_code="ILC_DI04", params=["total", "med_e"]).run()


def test_correct_params_and_dataset_code(caplog):
    """This function is designed to test the accuracy of the data retrieval feature in a program.
    Specifically, it tests to ensure that the program returns a non-empty DataFrame when a correct
    dataset code is provided with correct params. The function is intended to be used in software
    development to verify that the program is correctly retrieving data from the appropriate dataset.
    """

    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04", params={"hhtyp": "total", "indic_il": "med_e"}
    ).run()

    assert isinstance(task, pd.DataFrame)
    assert not task.empty
    assert caplog.text == ""


def task_correct_requested_columns(caplog):
    """This function is designed to test the accuracy of the data retrieval feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log error
    when provided with a correct dataset_code, correct params and correct requested_columns.
    The function is intended to be used in software development to verify that the program is correctly
    retrieving data from the appropriate dataset.
    """
    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04",
        params={"hhtyp": "total", "indic_il": "med_e"},
        requested_columns=["updated", "geo", "indicator"],
    )
    task.run()

    assert isinstance(task, pd.DataFrame)
    assert not task.empty
    assert caplog.text == ""
    assert list(task.columns) == task.needed_columns


def test_wrong_needed_columns_names(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log error
    when provided with a correct dataset_code, correct parameters, but incorrect names of requested columns.
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04",
        params={"hhtyp": "total", "indic_il": "med_e"},
        requested_columns=["updated1", "geo1", "indicator1"],
    )
    with pytest.raises(ValueError, match="Provided columns are not available!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        f"Name of the columns: 'updated1 | geo1 | indicator1' are not in DataFrame. Please check spelling!\n"
        in caplog.text
    )
    assert f"Available columns: geo | time | indicator | label | updated" in caplog.text


def test_wrong_params_and_wrong_requested_columns_names(caplog):
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log error
    when provided with a correct dataset_code, incorrect parameters and incorrect names of requested columns.
    Test should log errors only related with wrong params - we are trying to check if program will stop after
    params validation. The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    task = eurostat.EurostatToDF(
        dataset_code="ILC_DI04",
        params={"hhhtyp": "total", "indic_ilx": "med_e"},
        requested_columns=["updated1", "geo1", "indicator1"],
    )
    with pytest.raises(ValueError, match="DataFrame is empty!"):
        with caplog.at_level(logging.ERROR):
            task.run()
    assert (
        f"Parameters: 'hhhtyp | indic_ilx' are not in dataset. Please check your spelling!\n"
        in caplog.text
    )
    assert (
        f"Possible parameters: freq | hhtyp | indic_il | unit | geo | time"
        in caplog.text
    )


def test_requested_columns_not_in_list():
    """This function is designed to test the accuracy of the error logging feature in a program.
    Specifically, it tests to ensure that the program is able to correctly identify and log error
    when provided with a correct dataset_code, correct params but incorrect requested_columns structure
    (as single string instead of list with strings).
    The function is intended to be used in software development to identify correct type errors
    and messages in the program's handling of codes.
    """
    with pytest.raises(
        TypeError, match="Requested columns should be provided as list of strings."
    ):
        eurostat.EurostatToDF(
            dataset_code="ILC_DI04",
            params={"hhtyp": "total", "indic_il": "med_e"},
            requested_columns="updated",
        ).run()
